"""
Returns X and Y coordinates (presumably longitude and
latitude), as well as a value for shapes from a shapefile.
The missing value separates coordinates between shapes.
"""

from __future__ import print_function

import numpy
import pyferret
import shapefile

def ferret_init(efid):
    """
    Initialization for the shapefile_readxyval PyEF
    """
    retdict = { "numargs": 3,
                "descript": "Returns X, Y, and a value from shapes in a shapefile.  "
                            "Missing value separates shapes.",
                "restype": pyferret.FLOAT_ARRAY,
                "axes": ( pyferret.AXIS_ABSTRACT,
                          pyferret.AXIS_ABSTRACT,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST, ),
                "argnames": ( "SHAPEFILE", "VALNAME", "MAXPTS", ),
                "argdescripts": ( "Shapefile name (any extension given is ignored)",
                                  "Name of value to retrieve",
                                  "Max. number of points to return (-1 for all, but reads shapefile twice)", ),
                "argtypes": ( pyferret.STRING_ONEVAL,
                              pyferret.STRING_ONEVAL,
                              pyferret.FLOAT_ONEVAL, ),
                "influences": ( (False, False, False, False, False, False),
                                (False, False, False, False, False, False),
                                (False, False, False, False, False, False), ),
              }
    return retdict


def ferret_result_limits(efid):
    """
    Abstract axis limits for the shapefile_readxyval PyEF
    """
    maxpts = pyferret.get_arg_one_val(efid, pyferret.ARG3)
    maxpts = int(maxpts)
    if maxpts == -1:
        shapefile_name = pyferret.get_arg_one_val(efid, pyferret.ARG1)
        sf = shapefile.Reader(shapefile_name)
        maxpts = 0
        for shp in sf.shapes():
            maxpts += len(shp.points) + 1
    elif maxpts < 1:
        raise ValueError("MAXPTS must be a positive integer or -1")
    return ( (1, maxpts), (1, 3), None, None, None, None, )


def ferret_compute(efid, result, resbdf, inputs, inpbdfs):
    """
    Read the shapefile named in inputs[0] and assign result[:,0,0,0]
    with the X coordinates, result[:,1,0,0] with the Y coordinates
    of the shapes contained in the shapefile.  The missing value,
    resbdf, is assigned as the coordinates of a point separating
    different shapes.  Also assigns result[:,2,0,0] with the value
    of the field named in inputs[1] associated with each shape.
    """
    # Initialize all results to the missing value flag to make it easier later on
    result[:,:,:,:,:,:] = resbdf

    # Open the shapefile for reading and read the metadata
    sf = shapefile.Reader(inputs[0])

    # Find the index of the desired field in the shapefile
    fieldname = inputs[1].strip()
    # No function currently in the shapefile module to do this, so a bit of a hack here
    # Each field in shapefile is a tuple (name, type, size, precision)
    for k in range(len(sf.fields)):
        if sf.fields[k][0] == fieldname:
            break
    else:
        print("Known fields (name, type, size, precision):")
        for field in sf.fields:
            if field[0] != 'DeletionFlag':
                print("    %s" % str(field))
        raise ValueError("No field with the name '%s' found" % fieldname)
    if sf.fields[0][0] == 'DeletionFlag':
        field_index = k - 1
    else:
        field_index = k

    # Retrieve the coordinates of the shapes
    num_shapes = 0
    try:
        pt_index = 0
        for shp in sf.shapes():
            num_shapes += 1
            for pt in shp.points:
                result[pt_index,:2,0,0,0,0] = pt[:2]
                pt_index += 1
            # missing value coordinates (already assigned) separating shapes
            pt_index += 1
    except IndexError:
        # hit the maximum number of points
        pass
    if num_shapes < 1:
        raise ValueError("No shapes found")

    # Retrieve the field values
    rec_index = 0
    for rec in sf.records():
        result[rec_index,2,0,0,0,0] = float(rec[field_index])
        rec_index += 1
        # only get field values for shapes that were read
        if rec_index >= num_shapes:
            break


#
# The rest of this is for testing from the command line
#
if __name__ == "__main__":
    import time

    # make sure ferret_init does not cause problems
    info = ferret_init(0)

    resbdf = numpy.array([-9999.0], dtype=numpy.float64)
    inpbdfs = numpy.array([-8888.0, -7777.0], dtype=numpy.float64)
    maxpts = 3200 * 2400
    result = -6666.0 * numpy.ones((maxpts,3,1,1,1,1), dtype=numpy.float64, order='F')
    print("ferret_compute start: time = %s" % time.asctime())
    # INTPTLAT10 == latitude of an internal point in each county
    ferret_compute(0, result, resbdf, ("tl_2010_us_county10", "INTPTLAT10", maxpts, ), inpbdfs)
    print("ferret_compute done; time = %s" % time.asctime())
    good_x = numpy.logical_and((-180.0 <= result[:,0,0,0,0,0]), (result[:,0,0,0,0,0] <= -65.0))
    good_x = numpy.logical_or(good_x,
                 numpy.logical_and((172.0 <= result[:,0,0,0,0,0]), (result[:,0,0,0,0,0] <= 180.0)))
    good_y = numpy.logical_and((17.0 <= result[:,1,0,0,0,0]), (result[:,1,0,0,0,0] <= 72.0))
    if numpy.logical_xor(good_x, good_y).any():
        raise ValueError("good_x != good_y")
    missing_x = ( result[:,0,0,0,0,0] == resbdf )
    if numpy.logical_xor(good_x, numpy.logical_not(missing_x)).any():
        raise ValueError("good_x != not missing_x")
    missing_y = ( result[:,1,0,0,0,0] == resbdf )
    if numpy.logical_xor(good_y, numpy.logical_not(missing_y)).any():
        raise ValueError("good_y != not missing_y")
    count = 0
    at_end = False
    shape_num = 0
    total = 0
    for k in range(result.shape[0]):
        if missing_x[k]:
            if count == 0:
                at_end = True
            else:
                # print "Count[%d] = %d" % (shape_num, count)
                shape_num += 1
                total += count + 1
                count = 0
        elif at_end:
            raise ValueError("More than one missing value between shapes")
        else:
            count += 1
    total += count
    good_val = numpy.logical_and((17.0 <= result[:,2,0,0,0,0]), (result[:,2,0,0,0,0] <= 72.0))
    missing_val = ( result[:,2,0,0,0,0] == resbdf )
    if numpy.logical_xor(good_val, numpy.logical_not(missing_val)).any():
        raise ValueError("good_val != not missing_val")
    num_good = len(result[:,2,0,0,0,0][good_val])
    if num_good != shape_num:
        raise ValueError("number of values: expected %d, found %d" % (shape_num, num_good))
    print("shapefile_readxyval: SUCCESS")




'''
A Python program for testing the regrid3d.CurvRect3DRegridder class "by hand"
and also serves as a coding example of using this class.

@author: Karl Smith
'''

from __future__ import print_function

import sys
import numpy
import ESMP
from esmpcontrol import ESMPControl
from regrid3d import CurvRect3DRegridder


GULF_MEX_LONS = (262.0, 278.0, 0.5)
GULF_MEX_LATS = (18.0, 31.0, 0.5)
# Gulf of Mexico Relief of the Surface of the Earth (meters) 98W:82W:0.5 (columns), 18N:31N:0.5 (rows)
GULF_MEX_ROSE = (
      ( 1627.1,  2130.9,  1317.9,   373.8,   125.7,    45.0,    25.0,    20.0,    25.0,    25.0,
          15.0,    15.0,    25.0,    25.0,    40.0,    74.3,   169.7,   179.7,   151.3,    23.0,
          -3.2, -1638.9, -3668.0, -4098.9, -3075.3, -3017.9, -4554.6, -4969.7, -4961.6, -5008.2,
       -4759.7, -4932.3, -4016.7, ),
      ( 1951.7,  1669.0,   685.7,    63.4,    50.7,   779.4,   400.0,   -70.7,   -49.0,   -16.2,
          -6.1,     2.0,    -9.6,     5.0,    69.7,    80.1,   194.7,   210.2,   110.6,     6.1,
          -0.3,  -192.5, -1587.3, -4357.8, -4401.9, -4397.7, -2980.1, -2035.6,  -463.5, -2265.7,
       -4933.4, -4959.4, -4490.4, ),
      ( 2220.5,  2654.2,  1232.2,    60.8,   -10.0,  -217.9, -1170.2,  -807.9,  -599.1,  -388.8,
        -101.1,   -40.1,   -14.0,    -8.1,    19.9,    65.0,   139.4,   175.8,   115.2,    40.3,
           0.0,  -220.1,  -919.6, -4284.3, -4502.4, -4474.9, -4418.9, -3648.0, -2258.9, -2581.0,
       -2998.4, -3578.1, -3493.9, ),
      ( 2593.1,  2368.9,  1097.8,    29.9,  -225.9, -1889.3, -2357.3, -1504.2, -1161.4,  -939.7,
        -702.9,  -141.0,   -47.9,   -29.0,   -13.1,    29.7,    94.8,   150.0,   110.0,    70.5,
           0.0,  -340.6, -1551.1, -4353.5, -4516.0, -4534.5, -4455.3, -4438.0, -4153.9, -3078.7,
       -3004.0, -3228.8, -3595.4, ),
      ( 1944.0,   923.6,   191.8,   -12.8, -1599.3, -2398.4, -2802.9, -1996.6, -1597.9, -1039.4,
       -1061.3, -1345.8,   -62.4,   -25.9,   -21.1,    -9.9,    79.4,   120.0,    65.3,    30.5,
          -1.0,  -278.2,  -994.7, -1794.1, -4315.4, -4438.8, -4437.1, -4443.0, -4378.4, -3605.0,
       -3977.6, -4239.2, -2402.8, ),
      (  459.9,   210.9,    -9.9,  -588.9, -1913.3, -2591.4, -3053.4, -2975.4, -2016.1, -1992.1,
       -1619.8, -1926.1,   -38.2,   -23.6,   -23.0,    -0.1,    30.0,    30.2,    24.8,    15.0,
          10.0,     5.0,  -262.5, -1178.6, -2949.8, -4621.6, -4421.9, -4450.7, -4401.8, -4361.9,
       -4401.3, -4437.3, -3876.6, ),
      (  108.9,    23.3,   -97.1, -1136.2, -2098.3, -2602.7, -3192.4, -3178.5, -2608.1, -2588.1,
       -2597.9,  -660.0,    -9.9,   -40.1,   -29.0,   -10.1,     5.0,    10.0,    10.0,    10.0,
          10.0,    10.0,     0.0,  -276.1, -1413.3, -4101.5, -3835.0, -4424.2, -4415.9, -4443.6,
       -4443.1, -4460.6, -4399.8, ),
      (   69.7,   -10.0,  -402.6, -1785.3, -2391.5, -2800.2, -3278.6, -3408.0, -3111.9, -3002.1,
       -3014.9,  -782.6,   -44.6,   -17.9,   -21.9,   -29.0,   -21.0,   -14.0,   -12.0,     0.0,
           0.0,   -10.0,   -10.0,  -103.2, -1775.1, -2001.9, -2965.9, -3711.9, -3032.2, -2801.0,
          -3.1,  -920.6, -1273.1, ),
      (   64.0,   -10.0, -1313.9, -2190.4, -2769.9, -2899.3, -3317.6, -3536.4, -3564.0, -3364.1,
       -3331.0, -1715.7,   -13.2,   -47.9,   -19.7,   -46.2,   -45.9,   -40.0,   -33.1,    -3.1,
         -24.9,    -4.2,   -10.0,  -202.8, -1409.1, -1982.8,  -298.8,     0.0,  -575.5,  -339.9,
          -3.9,   -10.0,   -10.0, ),
      (   -8.5,   -10.0, -1714.1, -2321.6, -2790.2, -3072.8, -3446.8, -3621.8, -3639.0, -3635.0,
       -3581.5, -3569.9, -3367.2, -1055.4,  -227.7,  -117.0,   -53.2,   -47.7,   -61.2,   -55.2,
         -53.0,   -50.0,  -133.0,  -507.9, -1065.2, -1840.9, -2201.4,  -907.3,   192.0,    32.1,
          -2.2,   -10.0,   -10.0, ),
      (    7.2,   -45.9, -1526.8, -2196.7, -2588.5, -3178.9, -3593.6, -3630.1, -3623.9, -3628.0,
       -3621.0, -3619.0, -3709.8, -3908.8, -3558.9, -1507.3,  -237.2,   -45.2,    -4.3,   -62.2,
         -48.1,   -32.1,  -394.2,  -807.6, -1193.3, -1139.0, -2075.7, -2395.8, -2128.4, -1034.6,
          42.3,   112.0,   125.4, ),
      (  134.6,   -39.1, -1078.0, -2179.4, -2595.1, -3002.4, -3565.4, -3618.1, -3666.8, -3679.0,
       -3648.2, -3598.0, -3597.9, -3602.0, -3695.7, -3695.5, -1106.2,  -725.6,   -12.7,   -10.0,
         -92.3,  -151.6,  -591.2,  -977.7, -2939.2, -1435.4, -3011.0, -2948.8, -2399.3, -2182.2,
       -1800.0, -1789.0, -1570.9, ),
      (   55.0,   -35.2,  -838.4, -1609.6, -2488.7, -3268.8, -3594.0, -3613.1, -3724.5, -3727.1,
       -3714.9, -3671.4, -3739.6, -3718.0, -3655.8, -3607.0, -3826.2, -3652.7, -3069.2,  -352.8,
        -181.9,  -396.3,  -994.3, -1254.2, -3228.1, -3401.8, -3468.6, -2829.9, -2184.4, -1017.2,
       -1006.6,  -808.0, -1059.1, ),
      (   -8.7,   -18.7,  -141.0, -1008.6, -2264.6, -3205.6, -3619.4, -3620.2, -3654.8, -3655.1,
       -3631.0, -3622.0, -3611.0, -3609.0, -3616.1, -3611.1, -3600.8, -3851.1, -3622.7, -2515.4,
       -1578.2, -1492.1, -1769.6, -3051.5, -3383.7, -3415.2, -3402.4, -3268.5, -1954.1,  -192.3,
         -40.4,   -13.9,   -34.0, ),
      (    2.1,   -10.0,   -65.5,  -487.2, -1596.1, -2703.7, -3579.7, -3627.0, -3628.0, -3612.1,
       -3609.9, -3578.6, -3542.4, -3511.9, -3587.7, -3583.2, -3545.1, -3585.9, -3544.5, -3302.6,
       -3388.8, -3189.2, -3241.8, -3287.6, -3321.9, -3363.9, -3249.2, -2189.4,  -127.6,   -33.0,
         -43.8,   -35.9,   -10.0, ),
      (   17.1,     2.0,   -25.7,  -129.9, -1088.0, -1432.3, -3386.1, -3450.0, -3524.9, -3534.9,
       -3478.5, -3322.9, -3172.3, -3173.0, -3485.8, -3474.6, -3377.6, -3338.1, -3346.9, -3323.2,
       -3332.8, -3348.0, -3326.2, -3248.5, -3205.8, -3292.8, -3125.2,  -576.5,   -10.0,   -95.3,
         -54.2,   -34.1,   -10.1, ),
      (   17.2,     5.1,   -26.6,   -16.6,  -984.4, -1684.9, -2579.5, -3005.0, -3172.7, -3123.2,
       -2726.5, -2209.1, -2458.3, -2407.9, -3268.3, -3300.9, -3207.9, -3135.3, -3060.5, -2991.1,
       -3029.2, -3116.5, -3162.7, -3184.2, -3205.8, -3252.9, -2952.0,  -322.8,   -16.8,  -100.2,
         -44.2,   -30.0,   -12.1, ),
      (   13.1,   -10.0,   -29.0,   -98.8, -1000.0, -1589.9, -1674.0, -1989.6, -1577.0, -1851.7,
       -1797.0, -1904.0, -1999.0, -2290.0, -2206.1, -2795.8, -2937.9, -2917.3, -2723.1, -2597.1,
       -2752.5, -2909.1, -2999.6, -3094.5, -3177.6, -3381.8, -2460.4,  -200.7,   -85.4,   -73.2,
         -36.1,   -16.1,     0.9, ),
      (   14.0,    -9.7,   -41.8,  -140.8,  -789.2, -1386.5, -1399.6, -1211.4, -1293.0, -1023.3,
       -1212.6, -1503.1, -1733.2, -1991.2, -1894.8, -1864.4, -2391.9, -2511.8, -2282.3, -2586.4,
       -2724.9, -2879.5, -2959.6, -3129.6, -3203.8, -3205.9,  -817.6,  -102.9,   -89.2,   -50.3,
         -33.1,    -9.1,     7.8, ),
      (   19.3,    -9.6,   -28.7,   -67.7,  -205.8,  -796.4,  -994.9,  -950.1,  -881.5,  -601.2,
        -796.9,  -869.5,  -814.3, -1010.9, -1083.3, -1199.0, -1203.9, -1873.3, -1800.9, -2018.7,
       -2598.2, -2898.5, -3047.9, -3258.1, -3224.9, -1994.5,  -406.1,  -116.8,   -65.2,   -39.1,
         -17.2,    13.8,    20.2, ),
      (   41.6,    14.1,    -5.8,   -28.8,   -43.9,   -61.6,   -58.8,   -89.9,   -80.0,  -100.0,
        -103.2,  -105.9,  -134.1,  -203.9,  -211.7,  -228.9,  -537.9, -1026.6, -1397.3, -2180.6,
       -2476.4, -2800.7, -2668.6, -2293.7, -1026.3,  -600.2,  -230.9,   -80.6,   -44.1,   -29.0,
          -9.1,     8.9,    38.4, ),
      (  105.0,    44.1,     7.2,     0.0,    -7.9,   -27.9,   -31.0,   -28.0,   -41.9,   -47.0,
         -47.0,   -49.0,   -49.0,   -39.0,   -29.0,   -37.9,  -196.6,  -395.4,  -885.8, -1677.7,
       -2276.7, -1953.6,  -873.2,  -570.8,  -303.7,  -200.0,  -100.1,   -50.1,   -33.1,   -23.1,
         -10.1,    36.6,    35.8, ),
      (   85.3,    81.1,    31.2,    15.0,    11.1,     2.1,    -7.0,   -17.9,   -19.0,   -20.0,
         -22.0,   -25.0,   -18.0,   -10.0,    -3.0,    -8.0,   -24.8,   -11.4,   -77.7,  -502.6,
       -1330.0, -1583.2,  -704.2,  -377.0,  -219.5,   -99.1,   -42.1,   -34.1,   -28.0,   -14.1,
          -4.1,    16.4,    25.0, ),
      (  189.7,    83.3,    79.0,    45.2,    29.9,    15.0,   -10.7,    -1.3,   -12.0,    -9.0,
         -12.0,    -9.0,    -1.0,    -1.0,     1.0,     1.0,    -0.3,    -9.0,   -13.9,   -42.7,
         -51.4,  -109.8,  -397.0,  -187.9,   -58.6,   -20.2,    -9.0,   -20.1,   -19.0,    -4.2,
          11.8,    24.0,    36.4, ),
      (  210.0,   148.8,   104.6,    76.4,    56.2,    26.2,    14.1,     8.1,     1.1,     2.9,
           5.0,     3.1,     6.0,     1.0,     3.0,     3.0,    -9.6,    -0.3,   -10.0,   -24.8,
         -28.1,   -28.0,   -95.3,   -58.0,   -31.1,     4.8,     8.9,     0.1,    -1.0,    16.0,
          12.0,    32.1,    24.7, ),
      (  259.9,   163.1,   111.3,    66.4,   105.2,    97.8,    20.5,    30.1,    20.2,    17.1,
          12.0,    14.0,     8.0,     7.0,     9.0,     6.1,    15.9,    20.1,     9.0,    14.8,
          -1.0,     4.2,    -0.8,     3.8,    17.9,    46.0,    53.6,    21.5,    39.1,    32.4,
          34.1,    31.9,    19.2, ),
      (  272.2,   149.6,   114.9,   130.2,    88.9,    55.0,    95.1,    92.3,    93.1,    58.1,
          52.1,    23.1,    11.9,    71.1,    66.6,    91.5,    70.5,    76.5,    59.7,    54.1,
          11.9,    68.2,    64.6,    76.9,    37.2,    37.3,    44.9,    63.2,    76.9,    67.6,
          51.2,    33.1,     9.2, ),
)


def createExampleCurvData():
    '''
    Creates and returns example longitude, latitudes, depth, and data
    for a curvilinear grid.  Assigns grid center point data[i,j,k]
        = -2 * sin(lon[i,j,k] + 20) * cos(lat[i,j,k]) / log(depth[i,j,k] + 10)
            for valid center points
        = 1.0E20 for invalid center points

    Arguments:
        None
    Returns:
        (corner_lons, corner_lats, corner_depths, corner_invalid,
         center_lons, center_lats, center_depths, center_invalid, data) where:
        corner_lons:    numpy 3D array of curvilinear corner point longitude coordinates
        corner_lats:    numpy 3D array of curvilinear corner point latitude coordinates
        corner_depths:  numpy 3D array of curvilinear corner point depth coordinates
        corner_invalid: numpy 3D array of logicals indicating invalid corner points
        center_lons:    numpy 3D array of curvilinear center point longitude coordinates
        center_lats:    numpy 3D array of curvilinear center point latitude coordinates
        center_depths:  numpy 3D array of curvilinear center point depth coordinates
        center_invalid: numpy 3D array of logicals indicating invalid center points
        data:           numpy 3D array of curvilinear center point data values
    '''
    # Longitudes for the Gulf of Mexico ROSE data
    start = GULF_MEX_LONS[0]
    end = GULF_MEX_LONS[1]
    delta = GULF_MEX_LONS[2]
    # numpy.arange does not include the ending value
    end += 0.5 * delta
    center_lons = numpy.arange(start, end, delta)
    corner_lons = numpy.arange(start - 0.5 * delta, end + 0.5 * delta, delta)

    # Latitudes for the Gulf of Mexico ROSE data
    start = GULF_MEX_LATS[0]
    end = GULF_MEX_LATS[1]
    delta = GULF_MEX_LATS[2]
    # numpy.arange does not include the ending value
    end += 0.5 * delta
    center_lats = numpy.arange(start, end, delta)
    corner_lats = numpy.arange(start - 0.5 * delta, end + 0.5 * delta, delta)

    # arbitrary depths as sigma coordinates
    center_sigmas = numpy.array(    (0.200, 0.030, 0.075, 0.150, 0.300, 0.450, 0.600, 0.800) )
    corner_sigmas = numpy.array( (0.015, 0.025, 0.053, 0.113, 0.225, 0.375, 0.525, 0.700, 0.900) )
    center_shape = (center_lons.shape[0], center_lats.shape[0], center_sigmas.shape[0])
    corner_shape = (corner_lons.shape[0], corner_lats.shape[0], corner_sigmas.shape[0])

    # Expand longitudes, latitudes, and sigmas to 3D arrays
    center_lons = numpy.tile(center_lons, center_shape[1] * center_shape[2]) \
                       .reshape(center_shape, order='F')
    center_lats = numpy.tile(numpy.repeat(center_lats, center_shape[0]),
                             center_shape[2]) \
                       .reshape(center_shape, order='F')
    center_sigmas = numpy.repeat(center_sigmas, center_shape[0] * center_shape[1]) \
                         .reshape(center_shape, order='F')
    corner_lons = numpy.tile(corner_lons, corner_shape[1] * corner_shape[2]) \
                       .reshape(corner_shape, order='F')
    corner_lats = numpy.tile(numpy.repeat(corner_lats, corner_shape[0]),
                             corner_shape[2]) \
                       .reshape(corner_shape, order='F')
    corner_sigmas = numpy.repeat(corner_sigmas, corner_shape[0] * corner_shape[1]) \
                         .reshape(corner_shape, order='F')

    # Add a zeta coordinate (sea swell above earth radius) - less than 1.0 meter
    zeta = -1.0 * numpy.sin(numpy.deg2rad(center_lons)) * numpy.cos(numpy.deg2rad(center_lats))

    # Convert sigma, zeta coordinates to center depths
    # Modify the depths so nothing more than 325 meters deep to better match rectilinear data
    gulf_mex_center_rose = numpy.array(GULF_MEX_ROSE).flatten()
    gulf_mex_center_rose[ (gulf_mex_center_rose < -325.0) ] = -325.0
    gulf_mex_center_rose = numpy.tile(gulf_mex_center_rose, center_shape[2]) \
                                .reshape(center_shape, order='F')
    center_invalid = ( gulf_mex_center_rose > -10.0 )
    center_depths = center_sigmas * (zeta - gulf_mex_center_rose) - zeta

    # Compute the data values for these center points
    insea = numpy.logical_not(center_invalid)
    data = numpy.empty(center_shape, dtype=numpy.float64, order='F')
    data[center_invalid] = 1.0E20
    data[insea] = -2.0 * numpy.sin(numpy.deg2rad(center_lons[insea] + 20.0)) \
                       * numpy.cos(numpy.deg2rad(center_lats[insea])) \
                       / numpy.log10(center_depths[insea] + 10.0)

    # Convert sigma, zeta coordinates to corner depths
    # Estimate the corner depths using the center depths;
    # edges not too critical since most are over land
    gulf_mex_corner_rose = numpy.zeros((corner_shape[0], corner_shape[1]),
                                       dtype=numpy.float64, order='F')
    gulf_mex_corner_rose[1:-1,1:-1] = ( gulf_mex_center_rose[:-1, :-1, 0] + \
                                        gulf_mex_center_rose[1:,  :-1, 0] + \
                                        gulf_mex_center_rose[1:,  1:,  0] + \
                                        gulf_mex_center_rose[:-1, 1:,  0] ) / 4.0
    gulf_mex_corner_rose[ 0, 1:-1] = 2.0 * gulf_mex_corner_rose[ 1, 1:-1] - gulf_mex_corner_rose[ 2, 1:-1]
    gulf_mex_corner_rose[-1, 1:-1] = 2.0 * gulf_mex_corner_rose[-2, 1:-1] - gulf_mex_corner_rose[-3, 1:-1]
    gulf_mex_corner_rose[:,  0] = 2.0 * gulf_mex_corner_rose[:,  1] - gulf_mex_corner_rose[:,  2]
    gulf_mex_corner_rose[:, -1] = 2.0 * gulf_mex_corner_rose[:, -2] - gulf_mex_corner_rose[:, -3]
    gulf_mex_corner_rose = numpy.tile(gulf_mex_corner_rose.flatten('F'), corner_shape[2]) \
                                .reshape(corner_shape, order='F')
    zeta = -1.0 * numpy.sin(numpy.deg2rad(corner_lons)) * numpy.cos(numpy.deg2rad(corner_lats))
    corner_invalid = ( gulf_mex_corner_rose > -10.0 )
    corner_depths = corner_sigmas * (zeta - gulf_mex_corner_rose) - zeta

    return (corner_lons, corner_lats, corner_depths, corner_invalid,
            center_lons, center_lats, center_depths, center_invalid, data)


def createExampleRectData():
    '''
    Creates and returns example longitude, latitudes, depth, and data
    for a rectilinear grid.  Covers approximately the same region given
    by createExampleCurvData.  Assigns grid center point data[i,j,k]
        = -2 * sin(lon[i] + 20) * cos(lat[j]) / log(depth[k] + 10)
            for valid center points
        = 1.0E34 for invalid center points

    Arguments:
        None
    Returns:
        (corner_lons, corner_lats, corner_depths,
         center_lons, center_lats, center_depths, data) where:
        corner_lons:    numpy 1D array of rectilinear corner plane longitudes
        corner_lats:    numpy 1D array of rectilinear corner plane latitudes
        corner_depths:  numpy 1D array of rectilinear corner plane depths
        corner_invalid: numpy 3D array of logicals indicating invalid corner points
        center_lons:    numpy 1D array of rectilienar center plane longitudes
        center_lats:    numpy 1D array of rectilinear center plane latitudes
        center_depths:  numpy 1D array of rectilinear center plane depths
        center_invalid: numpy 3D array of logicals indicating invalid center points
        data:           numpy 3D array of rectilinear center point data values
    '''
    # Longitudes for the Gulf of Mexico ROSE data
    start = GULF_MEX_LONS[0]
    end = GULF_MEX_LONS[1]
    delta = GULF_MEX_LONS[2]
    # numpy.arange does not include the ending value
    end += 0.5 * delta
    center_lons = numpy.arange(start, end, delta)
    corner_lons = numpy.arange(start - 0.5 * delta, end + 0.5 * delta, delta)
    # Latitudes for the Gulf of Mexico ROSE data
    start = GULF_MEX_LATS[0]
    end = GULF_MEX_LATS[1]
    delta = GULF_MEX_LATS[2]
    # numpy.arange does not include the ending value
    end += 0.5 * delta
    center_lats = numpy.arange(start, end, delta)
    corner_lats = numpy.arange(start - 0.5 * delta, end + 0.5 * delta, delta)
    # Arbitrarily chosen depths
    center_depths = numpy.array(   (10.0,  25.0,  50.0, 100.0, 150.0, 250.0) )
    # Edges for these irregularily spaced depths      
    corner_depths = numpy.array( (5.0,  17.5,  37.5,  75.0, 125.0, 200.0, 300.0) )

    # Create the data values
    center_shape = (center_lons.shape[0], center_lats.shape[0], center_depths.shape[0])
    center_3d_lons = numpy.tile(center_lons, center_shape[1] * center_shape[2]) \
                          .reshape(center_shape, order='F')
    center_3d_lats = numpy.tile(numpy.repeat(center_lats, center_shape[0]),
                                center_shape[2]) \
                          .reshape(center_shape, order='F')
    center_3d_depths = numpy.repeat(center_depths, center_shape[0] * center_shape[1]) \
                            .reshape(center_shape, order='F')
    data = -2.0 * numpy.sin(numpy.deg2rad(center_3d_lons + 20.0)) \
                * numpy.cos(numpy.deg2rad(center_3d_lats)) \
                / numpy.log10(center_3d_depths + 10.0)
           
    # Use GULD_MEX_ROSE as a land mask to mark undefined data
    gulf_mex_center_rose = numpy.array(GULF_MEX_ROSE).flatten()
    gulf_mex_center_rose = numpy.tile(gulf_mex_center_rose, center_shape[2]) \
                                .reshape(center_shape, order='F')
    center_invalid = numpy.logical_or( ( gulf_mex_center_rose > -10.0 ),
                                       ( center_3d_depths > -gulf_mex_center_rose ) )
    data[ center_invalid ] = 1.0E34

    # Convert sigma, zeta coordinates in corner depths
    # Estimate the corner depths using the center depths;
    # edges not too critical since most are over land
    corner_shape = (corner_lons.shape[0], corner_lats.shape[0], corner_depths.shape[0])
    corner_3d_depths = numpy.repeat(corner_depths, corner_shape[0] * corner_shape[1]) \
                            .reshape(corner_shape, order='F')
    gulf_mex_corner_rose = numpy.zeros((corner_shape[0], corner_shape[1]),
                                       dtype=numpy.float64, order='F')
    gulf_mex_corner_rose[1:-1,1:-1] = ( gulf_mex_center_rose[:-1, :-1, 0] + \
                                        gulf_mex_center_rose[1:,  :-1, 0] + \
                                        gulf_mex_center_rose[1:,  1:,  0] + \
                                        gulf_mex_center_rose[:-1, 1:,  0] ) / 4.0
    gulf_mex_corner_rose[ 0, 1:-1] = 2.0 * gulf_mex_corner_rose[ 1, 1:-1] - gulf_mex_corner_rose[ 2, 1:-1]
    gulf_mex_corner_rose[-1, 1:-1] = 2.0 * gulf_mex_corner_rose[-2, 1:-1] - gulf_mex_corner_rose[-3, 1:-1]
    gulf_mex_corner_rose[:,  0] = 2.0 * gulf_mex_corner_rose[:,  1] - gulf_mex_corner_rose[:,  2]
    gulf_mex_corner_rose[:, -1] = 2.0 * gulf_mex_corner_rose[:, -2] - gulf_mex_corner_rose[:, -3]
    gulf_mex_corner_rose = numpy.tile(gulf_mex_corner_rose.flatten('F'), corner_shape[2]) \
                                .reshape(corner_shape, order='F')
    corner_invalid = numpy.logical_or( ( gulf_mex_corner_rose > -10.0 ),
                                       ( corner_3d_depths > -gulf_mex_corner_rose ) )

    return (corner_lons, corner_lats, corner_depths, corner_invalid,
            center_lons, center_lats, center_depths, center_invalid, data)


def printDiffs(grid_lons, grid_lats, grid_depths, undef_val, max_negl,
                 expect_data, found_data):
    '''
    Prints significant differences between expect_data and found_data
    along with the location of these differences

    Arguments:
        grid_lons:   numpy 3D array of grid longitudes
        grid_lats:   numpy 3D array of grid latitudes
        grid_depths: numpy 3D array of grid depths
        undef_val:   numpy array of one value; the undefined data value
        max_negl:    maximumum negligible absolute difference
        expect_data: numpy 3D array of expected data values
        found_data:  numpy 3D array of data values to check
    Returns:
        None
    Raises:
        ValueError:  if the array shapes do not match
    '''
    if len(grid_lons.shape) != 3:
        raise ValueError("grid_lons is not 3D")
    if grid_lats.shape != grid_lons.shape:
        raise ValueError("grid_lats.shape != grid_lons.shape")
    if grid_depths.shape != grid_lons.shape:
        raise ValueError("grid_depth.shape != grid_lons.shape")
    if expect_data.shape != grid_lons.shape:
        raise ValueError("expect_data.shape != grid_lons.shape")
    if found_data.shape != grid_lons.shape:
        raise ValueError("found_data.shape != grid_lons.shape")
    different = ( numpy.abs(expect_data - found_data) > max_negl )
    diff_lons = grid_lons[different]
    diff_lats = grid_lats[different]
    diff_depths = grid_depths[different]
    diff_expect = expect_data[different]
    diff_found = found_data[different]
    diff_list = [ ]
    for (lon, lat, depth, expect, found) in \
            zip(diff_lons, diff_lats, diff_depths, diff_expect, diff_found):
        if expect == undef_val:
            # most serious - should have been masked out
            diff_list.append([2, lon, lat, depth, expect, found])
        elif found == undef_val:
            # least serious - destination not covered by source
            diff_list.append([0, lon, lat, depth, expect, found])
        else:
            # might be of concern
            diff_list.append([1, lon, lat, depth, expect, found])
    # order primarily from least to most serious, 
    # secondarily smallest to largest longitude,
    # thirdly smallest to largest latitude
    # fourthly highest to deepest
    diff_list.sort()
    num_not_undef = 0
    num_undef = 0
    num_diff = 0
    for (_, lon, lat, depth, expect, found) in diff_list:
        if expect == undef_val:
            num_not_undef += 1
            print("lon = %#7.3f, lat = %7.3f, depth = %7.2f, expect =  undef, " \
                  "found = %#6.3f" % (lon, lat, depth, found))
        elif found == undef_val:
            num_undef += 1
            # print "lon = %#7.3f, lat = %7.3f, depth = %7.2f, expect = %#6.3f, " \
            #       "found =  undef" % (lon, lat, depth, expect)
        else:
            num_diff += 1
            print("lon = %#7.3f, lat = %7.3f, depth = %7.2f, expect = %#6.3f, " \
                  "found = %#6.3f, diff = %#6.3f" \
                  % (lon, lat, depth, expect, found, found - expect))
    print("%3d undefined when defined might be expected" % num_undef)
    print("%3d with absolute difference > %#.3f" % (num_diff, max_negl))
    print("%3d defined when undefined expected" % num_not_undef)
    print("%3d values in the grid" \
            % (expect_data.shape[0] * expect_data.shape[1] * expect_data.shape[2]))


if __name__ == '__main__':
    try:
        while True:
            print('cw2r: curvilinear with corners to rectilinear')
            print('co2r: curvilinear without corners to rectilinear')
            print('r2cw: rectilinear to curvilinear with corners')
            print('r2co: rectilinear to curvilinear without corners')
            print('Ctrl-D to quit')
            if sys.version_info[0] > 2:
                direction = input('Regrid test to run? ')
            else:
                direction = raw_input('Regrid test to run? ')
            direction = direction.strip().lower()
            if direction in ('cw2r', 'co2r', 'r2cw', 'r2co'):
                break
    except EOFError:
        raise SystemExit(0)

    # Synthesize curvilinear test data
    (curv_corner_lons, curv_corner_lats, curv_corner_depths, curv_corner_ignore,
     curv_center_lons, curv_center_lats, curv_center_depths, curv_center_ignore,
     curv_data) = createExampleCurvData()

    # Synthesize rectilinear test data
    (rect_corner_lons, rect_corner_lats, rect_corner_depths, rect_corner_ignore,
     rect_center_lons, rect_center_lats, rect_center_depths, rect_center_ignore,
     rect_data) = createExampleRectData()

    undef_val = numpy.array([-1.0E10], dtype=numpy.float64)

    # Create the expected results on the curvilinear grid
    curv_expect_data = curv_data.copy()
    curv_expect_data[curv_center_ignore] = undef_val

    # Create the expected results on the rectilinear grid
    rect_expect_data = rect_data.copy()
    rect_expect_data[rect_center_ignore] = undef_val

    # Initialize ESMP
    if not ESMPControl().startCheckESMP():
        raise RuntimeError("Unexpected failure to start ESMP")

    # Create the regridder
    regridder = CurvRect3DRegridder()

    if direction in ('cw2r', 'r2cw'):
        # Create the curvilinear grid with corner and center points
        regridder.createCurvGrid(curv_center_lons, curv_center_lats, curv_center_depths,
                                 curv_center_ignore, True, curv_corner_lons, curv_corner_lats,
                                 curv_corner_depths, curv_corner_ignore)
    elif direction in ('co2r', 'r2co'):
        # Create the curvilinear grid with only center points
        regridder.createCurvGrid(curv_center_lons, curv_center_lats, curv_center_depths,
                                 curv_center_ignore, True)
    else:
        raise ValueError("unexpected direction of %s" % direction)

    # Create the rectilinear grid with corner and center points
    regridder.createRectGrid(rect_center_lons, rect_center_lats, rect_center_depths,
                             rect_center_ignore, True, rect_corner_lons, rect_corner_lats,
                             rect_corner_depths, rect_corner_ignore)

    if direction in ('cw2r', 'co2r'):

        print("")
        if direction == 'cw2r':
            print("Examining rectilinear results from curvilinear with corners")
        else:
            print("Examining rectilinear results from curvilinear without corners")

        # Create the curvilinear source field
        regridder.assignCurvField(curv_data)

        # Create the rectilinear destination field
        regridder.assignRectField()

        # Generate the 3D rectilinear longitude, latitude, and depth arrays
        # only to simplify printing differences; not used in regridding
        rect_center_shape = (rect_center_lons.shape[0], rect_center_lats.shape[0],
                             rect_center_depths.shape[0])
        rect_3d_center_lons = numpy.tile(rect_center_lons,
                                         rect_center_shape[1] * rect_center_shape[2]) \
                                   .reshape(rect_center_shape, order='F')
        rect_3d_center_lats = numpy.tile(numpy.repeat(rect_center_lats,
                                                      rect_center_shape[0]),
                                         rect_center_shape[2]) \
                                   .reshape(rect_center_shape, order='F')
        rect_3d_center_depths = numpy.repeat(rect_center_depths,
                                             rect_center_shape[0] * rect_center_shape[1]) \
                                     .reshape(rect_center_shape, order='F')

        if direction == 'cw2r':
            # Regrid from curvilinear to rectilinear using the conserve method
            # Corners required for this method
            rect_regrid_data = regridder.regridCurvToRect(undef_val, ESMP.ESMP_REGRIDMETHOD_CONSERVE)
            # Print the differences between the expected and regrid data
            print("")
            print("analytic (expect) versus conserve regridded (found) differences")
            printDiffs(rect_3d_center_lons,
                       rect_3d_center_lats,
                       rect_3d_center_depths,
                       undef_val, 0.001,
                       rect_expect_data,
                       rect_regrid_data)

        # Regrid from curvilinear to rectilinear using the bilinear method
        rect_regrid_data = regridder.regridCurvToRect(undef_val, ESMP.ESMP_REGRIDMETHOD_BILINEAR)
        # Print the differences between the expected and regrid data
        print("")
        print("analytic (expect) versus bilinear regridded (found) differences")
        printDiffs(rect_3d_center_lons,
                   rect_3d_center_lats,
                   rect_3d_center_depths,
                   undef_val, 0.12,
                   rect_expect_data,
                   rect_regrid_data)

        # Regrid from curvilinear to rectilinear using the patch method
        rect_regrid_data = regridder.regridCurvToRect(undef_val, ESMP.ESMP_REGRIDMETHOD_PATCH)
        # Print the differences between the expected and regrid data
        print("")
        print("analytic (expect) versus patch regridded (found) differences")
        printDiffs(rect_3d_center_lons,
                   rect_3d_center_lats,
                   rect_3d_center_depths,
                   undef_val, 1.0, 
                   rect_expect_data,
                   rect_regrid_data)

    elif direction in ('r2cw', 'r2co'):

        print("")
        if direction == 'r2cw':
            print("Examining curvilinear with corners results from rectilinear")
        else:
            print("Examining curvilinear without corners results from rectilinear")

        # Create the rectilinear source field
        regridder.assignRectField(rect_data)

        # Create the curvilinear destination field
        regridder.assignCurvField(None)

        if direction == 'r2cw':
            # Regrid from rectilinear to curvilinear using the conserve method
            # Corners required for this method
            curv_regrid_data = regridder.regridRectToCurv(undef_val, ESMP.ESMP_REGRIDMETHOD_CONSERVE)
            # Print the differences between the expected and regrid data
            print("")
            print("analytic (expect) versus conserve regridded (found) differences")
            printDiffs(curv_center_lons[:,:,1:-1],
                       curv_center_lats[:,:,1:-1],
                       curv_center_depths[:,:,1:-1],
                       undef_val, 0.001,
                       curv_expect_data[:,:,1:-1],
                       curv_regrid_data[:,:,1:-1])

        # Regrid from rectilinear to curvilinear using the bilinear method
        curv_regrid_data = regridder.regridRectToCurv(undef_val, ESMP.ESMP_REGRIDMETHOD_BILINEAR)
        # Print the differences between the expected and regrid data
        print("")
        print("analytic (expect) versus bilinear regridded (found) differences")
        printDiffs(curv_center_lons[:,:,1:-1],
                   curv_center_lats[:,:,1:-1],
                   curv_center_depths[:,:,1:-1],
                   undef_val, 0.25,
                   curv_expect_data[:,:,1:-1],
                   curv_regrid_data[:,:,1:-1])

        # Regrid from rectilinear to curvilinear using the patch method
        curv_regrid_data = regridder.regridRectToCurv(undef_val, ESMP.ESMP_REGRIDMETHOD_PATCH)
        # Print the differences between the expected and regrid data
        print("")
        print("analytic (expect) versus patch regridded (found) differences")
        printDiffs(curv_center_lons[:,:,1:-1],
                   curv_center_lats[:,:,1:-1],
                   curv_center_depths[:,:,1:-1],
                   undef_val, 1.1,
                   curv_expect_data[:,:,1:-1],
                   curv_regrid_data[:,:,1:-1])

    else:
        raise ValueError("unexpected direction of %s" % direction)

    # Done with this regridder
    regridder.finalize()

    # Done with ESMP    
    ESMPControl().stopESMP(True)




'''
Tests of ESMPControl singleton class

@author: Karl Smith
'''
import unittest
from esmpcontrol import ESMPControl


class Test(unittest.TestCase):
    '''
    Tests of ESMPControl singleton class
    '''


    def test01ESMPControlNew(self):
        '''
        Test of ESMPControl constructor
        '''
        ctrl1 = ESMPControl()
        ctrl2 = ESMPControl()
        self.assertTrue(ctrl1 is ctrl2, "ESMPControl is not a singleton")


    def test02ESMPControlStartStop(self):
        '''
        Test of ESMPControl.startCheckESMP
        '''
        ctrl = ESMPControl()
        # calling stopESMP before calling startCheckESMP should not cause any problems
        ctrl.stopESMP(False)
        self.assertTrue(ctrl.startCheckESMP(),
                        "First call to startCheckESMP did not return True")
        self.assertTrue(ctrl.startCheckESMP(),
                        "Second call to startCheckESMP did not return True")
        ctrl.stopESMP(True)
        self.assertFalse(ctrl.startCheckESMP(),
                         "Call to startCheckESMP after stopESMP did not return False")
        # another call to stopESMP should not cause any problems
        ctrl.stopESMP(False)
        # creating another instance should have no effect
        ctrl2 = ESMPControl()
        self.assertFalse(ctrl2.startCheckESMP(),
                         "Call to startCheckESMP from a 'new' instance after stopESMP did not return False")


if __name__ == "__main__":
    unittest.main()



"""
Convenience methods for executing common Ferret commands.
These method generate appropriate Ferret command strings and then
execute them using the pyferret.run method.
"""

import numbers
import pyferret


def setwindow(num=1, plotasp=None, axisasp=None, color=None, pal=None, 
              thick=None, logo=None, outline=None):
    """
    Assigns the plot window to use for subsequent plotting commands.
    Also provides assignment of common window plots.  
    Note that plotasp and axisasp cannot both be given.
        num (int): window number 1-8 to use for plots.
        plotasp (float): aspect ratio (Y/X) for the plot window.
            If not given, the current ratio is unchanged.
            The default ratio on start-up is 0.86
        axisasp (float): aspect ratio (Y/X) for the plot axes.
            If not given, the current ratio is unchanged.
            The default ratio on start-up is 0.75
        color (string, tuple of int): background color for the plot;
            can be one of the color names 'black', 'blue', 'green', 
            'lightblue', 'purple', or 'red', or a tuple
            of int values in [0,100] giving RGB or RGBA values.
            If not given, the current value is unchanged.
            The default background color on start-up is opaque white.
        pal (string): default color palette to use in plots.
            If not given, thr current value is unchanged.
        thick (float): line thickness scaling factor for the plot.
            If not given, the current scaling factor is unchanged.
            The default line thickness scaling factor on start-up is 1.0
        logo (boolean): include the Ferret logo in the plot?
            If not given, the current value is unchanged.
            The default on start-up is to include the logo.
        outline (float): if positive, thickness of polygon outlines;
            used to fix the 'thin white line' issue in plots.
            If not given, the current value is unchanged.
            The default on start-up is zero (no outlines drawn).
            
    Raises a ValueError if a problem occurs.
    """
    # create and execute the SET WINDOW command
    cmdstr = 'SET WINDOW'
    if (plotasp is not None) and (axisasp is not None):
        raise ValueError('only one of plotasp and axisasp can be given')
    if plotasp is not None:
        if (not isinstance(plotasp, numbers.Real)) or (plotasp <= 0):
            raise ValueError('plotasp, if given, must be a positive number')
        cmdstr += '/ASPECT=' + str(plotasp)
    if axisasp is not None:
        if (not isinstance(axisasp, numbers.Real)) or (axisasp <= 0):
            raise ValueError('axisasp, if given, must be a positive number')
        cmdstr += '/ASPECT=' + str(axisasp) + ':AXIS'
    if thick is not None:
        if (not isinstance(thick, numbers.Real)) or (thick <= 0):
            raise ValueError('thick, if given, must be a positive number')
        cmdstr += '/THICK=' + str(thick)
    if outline is not None:
        if (not isinstance(outline, numbers.Real)) or (outline < 0):
            raise ValueErrror('outline, if given, must be a non-negative number')
        cmdstr += '/OUTLINE=' + str(outline)
    if color is not None:
        if isinstance(color, str):
            cmdstr += '/COLOR=' + color
        elif isinstance(color, tuple):
            if (len(color) < 3) or (len(color) > 4):
                raise ValueError('a color tuple must have three or four integer values')
            cmdstr += '/COLOR=' + str(color)
        else:
            raise ValueError('given color %s is not a string or tuple' % str(color))
    if (not isinstance(num, numbers.Integral)) or (num <= 0) or (num > 8):
        raise ValueError('window number %s is not a integer in [1,8]' % str(num))
    cmdstr += ' ' + str(num)
    (errval, errmsg) = pyferret.run(cmdstr)
    if errval != pyferret.FERR_OK:
        raise ValueError('Problems executing Ferret command %s: %s' % (cmdstr, errmsg))
    if pal is not None:
        cmdstr = 'PALETTE ' + str(pal)
        (errval, errmsg) = pyferret.run(cmdstr)
        if errval != pyferret.FERR_OK:
            raise ValueError('Problems executing Ferret command %s: %s' % (cmdstr, errmsg))
    # create and execute the mode logo command if logo is given
    if logo is not None:
        if logo:
            cmdstr = 'SET MODE LOGO'
        else:
            cmdstr = 'CANCEL MODE LOGO'
        (errval, errmsg) = pyferret.run(cmdstr)
        if errval != pyferret.FERR_OK:
            raise ValueError('Problems executing Ferret command %s: %s' % (cmdstr, errmsg))


def settextstyle(font='', color='', bold = False, italic=False):
    """
    Sets the text style for any text in plots generated after this command
    using the Ferret SET TEXT command.
        font (string): name of the font to use; if empty, 'Arial' is used.
        color (string): color name, RGB tuple, or RGBA tuple describing the 
            color of the text.  The R,G,B, and A components are integer
            percentages; thus values in [0,100]
        bold (bool): use bold font?
        italic (bool): use italic font?
    """
    # First run CANCEL TEXT to clear any /BOLD and /ITALIC
    (errval, errmsg) = pyferret.run('CANCEL TEXT/ALL')
    if errval != pyferret.FERR_OK:
        raise ValueError('problems resetting text style to default: %s' % errmsg)
    # Now run SET TEXT with the appropriate qualifiers
    cmdstr = 'SET TEXT'
    if font:
        cmdstr += '/FONT='
        cmdstr += font
    else:
        cmdstr += '/FONT=Arial'
    if color:
        cmdstr += '/COLOR=' + str(color)
    if bold:
        cmdstr += '/BOLD'
    if italic:
        cmdstr += '/ITALIC'
    (errval, errmsg) = pyferret.run(cmdstr)
    if errval != pyferret.FERR_OK:
        raise ValueError('problems setting text style (%s): %s' % (cmdstr, errmsg))


def showdata(brief=True, qual=''):
    """
    Show the Ferret information about all datasets currently open in Ferret.  
    This uses the Ferret SHOW DATA command to create and display the information.
        brief (boolean): if True (default), a brief report is shown;
            otherwise a full report is shown.
        qual (string): Ferret qualifiers to add to the SHOW DATA command
    """
    if not isinstance(qual, str):
        raise ValueError('qual (Ferret qualifiers) must be a string')
    cmdstr = 'SHOW DATA'
    if not brief:
        cmdstr += '/FULL'
    if qual:
        cmdstr += qual
    (errval, errmsg) = pyferret.run(cmdstr)
    if errval != pyferret.FERR_OK:
        raise ValueError('Ferret command "%s" failed: %s' % (cmdstr, errmsg))


def contourplot(fvar, region=None, over=False, pal=None, qual=''):
    """
    Create a contour plot of the specified Ferret variable using the Ferret CONTOUR command.
    Using the fill method to generated a color-filled contour plot.
    The variable needs to be 2D (or qualifiers need to be added to specify a 2D slice).
        fvar (string or FerVar): Ferret variable to plot
        region (FerRegion): space-time region to plot; 
                if None, the full extents of the data will be used
        over (bool): overlay on an existing plot?
        pal (string): color palette to use
        qual (string): qualifiers to add to the Ferret SHADE command
    """
    if not isinstance(qual, str):
        raise ValueError('qual (Ferret qualifiers) must be a string')
    if isinstance(fvar, str):
        plotvar = fvar
    elif isinstance(fvar, pyferret.FerVar):
        plotvar = fvar._definition
    else:
        raise ValueError('fvar (Ferret variable to plot) must be a string or FerVar')
    cmdstr = 'CONTOUR'
    if over:
        cmdstr += '/OVER'
    if region is not None:
        if not isinstance(region, pyferret.FerRegion):
            raise ValueError('region, if given, must be a FerRegion')
        cmdstr += region._ferretqualifierstr();
    if pal is not None:
        cmdstr += '/PALETTE=' + str(pal)
    if qual:
        cmdstr += qual
    cmdstr += ' '
    cmdstr += plotvar
    (errval, errmsg) = pyferret.run(cmdstr)
    if errval != pyferret.FERR_OK:
        raise ValueError('Ferret shade command (%s) failed: %s' % (cmdstr, errmsg))


def fillplot(fvar, region=None, line=False, over=False, pal=None, qual=''):
    """
    Create a color-filled contour plot of the specified Ferret variable using the Ferret 
    FILL command.  Drawing of the contour lines themselves is optional.
    The variable needs to be 2D (or qualifiers need to be added to specify a 2D slice).
        fvar (string or FerVar): Ferret variable to plot
        region (FerRegion): space-time region to plot; 
                if None, the full extents of the data will be used
        line (bool): draw the contour lines?
        over (bool): overlay on an existing plot?
        pal (string): color palette to use
        qual (string): qualifiers to add to the Ferret SHADE command
    """
    if not isinstance(qual, str):
        raise ValueError('qual (Ferret qualifiers) must be a string')
    if isinstance(fvar, str):
        plotvar = fvar
    elif isinstance(fvar, pyferret.FerVar):
        plotvar = fvar._definition
    else:
        raise ValueError('fvar (Ferret variable to plot) must be a string or FerVar')
    cmdstr = 'FILL'
    if line:
        cmdstr += '/LINE'
    if over:
        cmdstr += '/OVER'
    if region is not None:
        if not isinstance(region, pyferret.FerRegion):
            raise ValueError('region, if given, must be a FerRegion')
        cmdstr += region._ferretqualifierstr();
    if pal is not None:
        cmdstr += '/PALETTE=' + str(pal)
    if qual:
        cmdstr += qual
    cmdstr += ' '
    cmdstr += plotvar
    (errval, errmsg) = pyferret.run(cmdstr)
    if errval != pyferret.FERR_OK:
        raise ValueError('Ferret shade command (%s) failed: %s' % (cmdstr, errmsg))


def shadeplot(fvar, region=None, over=False, pal=None, qual=''):
    """
    Create a colored plot of the specified Ferret variable using the Ferret SHADE command.
    (Plot coloring grid cells based on the variable value in that cell.)
    The variable needs to be 2D (or qualifiers need to be added to specify a 2D slice).
        fvar (string or FerVar): Ferret variable to plot
        region (FerRegion): space-time region to plot; 
                if None, the full extents of the data will be used
        over (bool): overlay on an existing plot?
        pal (string): color palette to use
        qual (string): qualifiers to add to the Ferret SHADE command
    """
    if not isinstance(qual, str):
        raise ValueError('qual (Ferret qualifiers) must be a string')
    if isinstance(fvar, str):
        plotvar = fvar
    elif isinstance(fvar, pyferret.FerVar):
        plotvar = fvar._definition
    else:
        raise ValueError('fvar (Ferret variable to plot) must be a string or FerVar')
    cmdstr = 'SHADE'
    if over:
        cmdstr += '/OVER'
    if region is not None:
        if not isinstance(region, pyferret.FerRegion):
            raise ValueError('region, if given, must be a FerRegion')
        cmdstr += region._ferretqualifierstr();
    if pal is not None:
        cmdstr += '/PALETTE=' + str(pal)
    if qual:
        cmdstr += qual
    cmdstr += ' '
    cmdstr += plotvar
    (errval, errmsg) = pyferret.run(cmdstr)
    if errval != pyferret.FERR_OK:
        raise ValueError('Ferret shade command (%s) failed: %s' % (cmdstr, errmsg))


def shadeland(res=20, color='gray', over=True, solid=True, X=None, Y=None):
    """
    Shades land masses for the current longitude-latitude plot or the specified X-Y region.
        res (int): ETOPO dataset resolution (in minutes of a degree) to use; 
            the corresponding ETOPO dataset (eg, etopo20.cdf for 20) must be available.
            Typically 5, 10, 20, 40, 60, 120 are available from Ferret's standard datasets.
        color (str): name of the color or color palette to used for land.
        over (bool): if true, overlay onto the current longitude-latitude plot;
            if False, create a new plot of the given region
        solid (bool): if True, shade the land in a single solid color;
            if False, shade different elevations using the given color palette
        X (str): longitude limits for the region as low:high
            If not given and over is False, '0E:360E' is used.
            if not given and over is True, the full range of the given plot is used.
        Y (str): latitude limits for the region as low:high
            If not given and over is False, '90S:90N' is used.
            If not given and over is True, the full range of the given plot is used.
    """
    cmdstr = 'GO fland'
    cmdstr += ' ' + str(res)
    cmdstr += ' ' + str(color)
    if over:
        cmdstr += ' OVERLAY'
    else:
        cmdstr += ' BASEMAP'
    if solid:
        cmdstr += ' SOLID'
    else:
        cmdstr += ' DETAILED'
    if X is not None:
        cmdstr += ' X=' + str(X)
    elif not over:
        # assign the default here even though this matches the script
        cmdstr += ' X=0E:360E'
    elif Y is not None:
        # if Y is given, then have to have an X argument;
        # needs to be a double wrap for a complete overlay
        cmdstr += ' X=0E:720E'
    if Y is not None:
        cmdstr += ' Y=' + str(Y)
    elif not over:
        # assign the default here even though this matches the script
        cmdstr += ' Y=90S:90N'
    (errval, errmsg) = pyferret.run(cmdstr)
    if errval != pyferret.FERR_OK:
        raise ValueError('Ferret script command (%s) failed: %s' % (cmdstr, errmsg))


def shadewater(res=20, color='gray', over=True, solid=True, X=None, Y=None):
    """
    Shades water masses for the current longitude-latitude plot or the specified region.
        res (int): ETOPO dataset resolution (in minutes of a degree) to use; 
            the corresponding ETOPO dataset (eg, etopo20.cdf for 20) must be available.
            Typically 5, 10, 20, 40, 60, 120 are available from Ferret's standard datasets.
        color (str): name of the color or color palette to used for water masses
        over (bool): if true, overlay onto the current longitude-latitude plot;
            if False, create a new plot of the given region
        solid (bool): if True, shade the water masses in a single solid color;
            if False, shade different depths using the given color palette
        X (str): longitude limits for the region as low:high; 
            if not given and over is False, '0E:360E' is used
        Y (str): latitude limits for the region as low:high; 
            if not given and over is False, '90S:90N'
    """
    cmdstr = 'GO focean'
    cmdstr += ' ' + str(res)
    cmdstr += ' ' + str(color)
    if over:
        cmdstr += ' OVERLAY'
    else:
        cmdstr += ' BASEMAP'
    if solid:
        cmdstr += ' SOLID'
    else:
        cmdstr += ' DETAILED'
    if X is not None:
        cmdstr += ' X=' + str(X)
    elif not over:
        # assign the default here even though this matches the script
        cmdstr += ' X=0E:360E'
    elif Y is not None:
        # if Y is given, then have to have an X argument;
        # needs to be a double wrap for a complete overlay
        cmdstr += ' X=0E:720E'
    if Y is not None:
        cmdstr += ' Y=' + str(Y)
    elif not over:
        # assign the default here even though this matches the script
        cmdstr += ' Y=90S:90N'
    (errval, errmsg) = pyferret.run(cmdstr)
    if errval != pyferret.FERR_OK:
        raise ValueError('Ferret script command (%s) failed: %s' % (cmdstr, errmsg))


def pointplot(fvar, vs=None, color=None, sym=None, symsize=None, thick=None,
              line=False, title=None, region=None, over=False, label=True, qual=''):
    """
    Create a point plot of the given value, or the given value versus another value 
    (if vs is given), possibly colored by another value (if color is a FerVar).
    To create a line plot with symbols, use the pointplot command with the line 
    option set to True.
        fvar (string or FerVar): Ferret variable to plot
        vs  (string or FerVar): if given, plot the above variable versus this variables
        color: line color or variable used to determine line color; if
            None: Ferret default color used,
            color name (string): name of color to use,
            color tuple (3 or 4-tupe of [0,100] int values): RGB or RGBA of color to use,
            FerVar or variable name string: color according to the value of this variable
                Note: color name strings are limited to (case insensitive) 
                      'black', 'red', 'green', 'blue', 'lightblue', 'purple'
                      other strings are assumed to be variable names
        sym (int): Ferret symbol number of the symbol to draw for the points.
                If not given, Ferret selects an appropriate symbol.
        symsize (float): size of the symbol in inches.
                If not given, Ferret select an appropriate size.
        thick (float): line thickness scaling factor when drawing symbols and lines
        line (bool): if True, draw a line between symbols/points
        title (string): title for the plot; if not given,  Ferret's default title is used
        region (FerRegion): space-time region to plot; 
                if None, the full extents of the data will be used
        over (bool): overlay onto an existing plot
        label (bool): if False, suppress all plot labels
        qual (string): qualifiers to add to the Ferret PLOT/LINE command
    """
    if not isinstance(qual, str):
        raise ValueError('qual (Ferret qualifiers) must be a string')
    if isinstance(fvar, str):
        plotvar = fvar
    elif isinstance(fvar, pyferret.FerVar):
        plotvar = fvar._definition
    else:
        raise ValueError('fvar (Ferret variable to plot) must be a string or FerVar')
    cmdstr = 'PLOT'
    if vs is not None:
        cmdstr += '/VS'
        plotvar += ','
        if isinstance(vs, str):
            plotvar += vs
        elif isinstance(vs, pyferret.FerVar):
            plotvar += vs._definition
        else:
            raise ValueError('vs (second Ferret variable to plot) must be a string or FerVar')
    if color is not None:
        if isinstance(color, tuple):
           cmdstr += '/COLOR=' + str(color)
        elif isinstance(color, pyferret.FerVar):
           cmdstr += '/RIBBON'
           plotvar += ',' + color._definition
        elif isinstance(color, str):
            if color.upper() in ('BLACK','RED','GREEN','BLUE','LIGHTBLUE','PURPLE'):
               cmdstr += '/COLOR=' + color
            else:
               cmdstr += '/RIBBON'
               plotvar += ',' + color
        else:
            raise ValueError('color must be a tuple, string, or FerVar')
    # always draw the symbols
    cmdstr += '/SYMBOL'
    if sym is not None:
        if (not isinstance(sym, numbers.Integral)) or (sym < 0) or (sym > 88):
            raise ValueError('sym is not a valid Ferret symbol number')
        if sym == 0:
            cmdstr += '=DOT'
        else:
            cmdstr += '=' + str(sym)
    if symsize is not None:
        if (not isinstance(symsize, numbers.Real)) or (symsize <= 0):
            raise ValueError('symsize must be a positive number')
        cmdstr += '/SIZE=' + str(symsize)
    if thick is not None:
        if (not isinstance(thick, numbers.Real)) or (thick <= 0):
            raise ValueError('thick must be a positive number')
        cmdstr += '/THICK=' + str(thick)
    if line:
        cmdstr += '/LINE'
    if title is not None:
       if not isinstance(title, str):
           raise ValueError('title must be a string')
       cmdstr += '/TITLE="' + title + '"'
    if over:
        cmdstr += '/OVER'
    if region is not None:
        if not isinstance(region, pyferret.FerRegion):
            raise ValueError('region, if given, must be a FerRegion')
        cmdstr += region._ferretqualifierstr();
    if not label:
        cmdstr += '/NOLABEL'
    if qual:
        cmdstr += qual
    cmdstr += ' '
    cmdstr += plotvar
    (errval, errmsg) = pyferret.run(cmdstr)
    if errval != pyferret.FERR_OK:
        raise ValueError('Ferret plot command (%s) failed: %s' % (cmdstr, errmsg))


def lineplot(fvar, vs=None, color=None, thick=None, dash=None, title=None, 
             region=None, along=None, over=False, label=True, qual=''):
    """
    Create a line plot of the given value, or the given value versus another value 
    (if vs is given), possibly colored by another value (if color is a FerVar).
    To create a line plot with symbols, use the pointplot command with the line 
    option set to True.
        fvar (string or FerVar): Ferret variable to plot
        vs  (string or FerVar): if given, plot the above variable versus this variables
        color: line color or variable used to determine line color; if
            None: Ferret default color used,
            color name (string): name of color to use,
            color tuple (3 or 4-tupe of [0,100] int values): RGB or RGBA of color to use,
            FerVar or variable name string: color according to the value of this variable
                Note: color name strings are limited to (case insensitive) 
                      'black', 'red', 'green', 'blue', 'lightblue', 'purple'
                      other strings are assumed to be variable names
        thick (float): line thickness scaling factor
        dash (4-tuple of float): draws the line as a dashed line where the four values 
             are the first drawn stroke length, first undrawn stroke length,
             second drawn stroke length, second undrawn stroke length of two dashes
        title (string): title for the plot; if not given,  Ferret's default title is used
        region (FerRegion): space-time region to plot; 
                if None, the full extents of the data will be used
        along (string; one of 'X','Y','Z','T','E','F', or lowercase): make a set of line 
                plots from two-dimensional data with this axis as the horizontal axis.
        over (bool): overlay onto an existing plot
        label (bool): if False, suppress all plot labels
        qual (string): qualifiers to add to the Ferret PLOT/LINE command
    """
    if not isinstance(qual, str):
        raise ValueError('qual (Ferret qualifiers) must be a string')
    if isinstance(fvar, str):
        plotvar = fvar
    elif isinstance(fvar, pyferret.FerVar):
        plotvar = fvar._definition
    else:
        raise ValueError('fvar (Ferret variable to plot) must be a string or FerVar')
    cmdstr = 'PLOT/LINE'
    if vs is not None:
        cmdstr += '/VS'
        plotvar += ','
        if isinstance(vs, str):
            plotvar += vs
        elif isinstance(vs, pyferret.FerVar):
            plotvar += vs._definition
        else:
            raise ValueError('vs (second Ferret variable to plot) must be a string or FerVar')
    if color is not None:
        if isinstance(color, tuple):
           cmdstr += '/COLOR=' + str(color)
        elif isinstance(color, pyferret.FerVar):
           cmdstr += '/RIBBON'
           plotvar += ',' + color._definition
        elif isinstance(color, str):
            if color.upper() in ('BLACK','RED','GREEN','BLUE','LIGHTBLUE','PURPLE'):
               cmdstr += '/COLOR=' + color
            else:
               cmdstr += '/RIBBON'
               plotvar += ',' + color
        else:
            raise ValueError('color must be a tuple, string, or FerVar')
    if thick is not None:
        if (not isinstance(thick, numbers.Real)) or (thick <= 0):
            raise ValueError('thick must be a positive number')
        cmdstr += '/THICK=' + str(thick)
    if dash is not None:
        if (not isinstance(dash, tuple)) or (len(dash) != 4):
            raise ValueError('dash must be a tuple of four floats');
        cmdstr += '/DASH=' + str(dash)
    if title is not None:
       if not isinstance(title, str):
           raise ValueError('title must be a string')
       cmdstr += '/TITLE="' + title + '"'
    if along is not None:
       axisnames = ('X','Y','Z','T','E','F','x','y','z','t','e','f')
       if not along in axisnames:
           raise ValueError('along must be one of ' + str(axisnames))
       cmdstr += '/ALONG=' + along.upper()
    if over:
        cmdstr += '/OVER'
    if region is not None:
        if not isinstance(region, pyferret.FerRegion):
            raise ValueError('region must be a FerRegion')
        cmdstr += region._ferretqualifierstr();
    if not label:
        cmdstr += '/NOLABEL'
    if qual:
        cmdstr += qual
    cmdstr += ' '
    cmdstr += plotvar
    (errval, errmsg) = pyferret.run(cmdstr)
    if errval != pyferret.FERR_OK:
        raise ValueError('Ferret plot command (%s) failed: %s' % (cmdstr, errmsg))


def saveplot(name, fmt='', xpix=None, ypix=None, xinch=None, yinch=None, qual=''):
    """
    Save the current plot.  If format is not given,
    the format is guessed from the filename extension.
        name (string): name of the file to contain the plot
        fmt (string): format of the plot file
        xpix (int): number of pixels in width of the saved raster (eg, PNG) plot
        ypix (int): number of pixels in the height of the saved raster (eg, PNG) plot
        xinch (float): inch width of the saved vector (eg, PDF) plot
        yinch (float): inch height of the save vector (eg, PDF) plot
        qual (string): qualifiers to add to the Ferret FRAME command
    """
    if not isinstance(name, str):
        raise ValueError('name (plot file name) must be a string')
    cmdstr = 'FRAME/FILE="%s"' % name
    if not isinstance(fmt, str):
        raise ValueError('fmt (plot file format) must be a string')
    if fmt:
        cmdstr += '/FORMAT=%s' % fmt
    if xpix is not None:
        if (not isinstance(xpix, int)) or (xpix <= 0):
            raise ValueError('xpix must be a positive integer')
        cmdstr += '/XPIX=' + str(xpix)
    if ypix is not None:
        if (not isinstance(ypix, int)) or (ypix <= 0):
            raise ValueError('ypix must be a positive integer')
        cmdstr += '/YPIX=' + str(ypix)
    if (xpix is not None) and (ypix is not None):
        raise ValueError('xpix and ypix cannot both be given')
    if xinch is not None:
        if (not isinstance(xinch, numbers.Real)) or (xinch <= 0.0):
            raise ValueError('xinch must be a positive number')
        cmdstr += '/XINCH=' + str(xinch)
    if yinch is not None:
        if (not isinstance(yinch, numbers.Real)) or (yinch <= 0.0):
            raise ValueError('yinch must be a positive number')
        cmdstr += '/YINCH=' + str(yinch)
    if (xinch is not None) and (yinch is not None):
        raise ValueError('xinch and yinch cannot both be given')
    if not isinstance(qual, str):
        raise ValueError('qual (Ferret qualifiers) must be a string')
    if qual:
        cmdstr += qual
    (errval, errmsg) = pyferret.run(cmdstr)
    if errval != pyferret.FERR_OK:
        raise ValueError('Ferret frame command (%s) failed: %s' % (cmdstr, errmsg))




"""
Returns the X and Y (presumably longitude,latitude)
coordinates from the points in the indicated shapefile.
The missing value separates coordinates between shapes.
"""

from __future__ import print_function

import numpy
import pyferret
import shapefile

def ferret_init(efid):
    """
    Initialization for the shapefile_readxy PyEF
    """
    retdict = { "numargs": 2,
                "descript": "Returns X,Y coordinates of shapes from shapefile.  "
                            "Missing value separates shapes.",
                "restype": pyferret.FLOAT_ARRAY,
                "axes": ( pyferret.AXIS_ABSTRACT,
                          pyferret.AXIS_ABSTRACT,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST, ),
                "argnames": ( "SHAPEFILE", "MAXPTS", ),
                "argdescripts": ( "Shapefile name (any extension given is ignored)",
                                  "Max. number of points to return (-1 for all, but reads shapefile twice)", ),
                "argtypes": ( pyferret.STRING_ONEVAL,
                              pyferret.FLOAT_ONEVAL, ),
                "influences": ( (False, False, False, False, False, False),
                                (False, False, False, False, False, False), ),
              }
    return retdict


def ferret_result_limits(efid):
    """
    Abstract axis limits for the shapefile_readxy PyEF
    """
    maxpts = pyferret.get_arg_one_val(efid, pyferret.ARG2)
    maxpts = int(maxpts)
    if maxpts == -1:
        shapefile_name = pyferret.get_arg_one_val(efid, pyferret.ARG1)
        sf = shapefile.Reader(shapefile_name)
        maxpts = 0
        for shp in sf.shapes():
            maxpts += len(shp.points) + 1
    elif maxpts < 1:
        raise ValueError("MAXPTS must be a positive integer or -1")
    return ( (1, maxpts), (1, 2), None, None, None, None, )


def ferret_compute(efid, result, resbdf, inputs, inpbdfs):
    """
    Read the shapefile named in inputs[0] and assign result[:,0,0,0]
    with the X coordinates, and result[:,1,0,0] with the Y coordinates
    of the shapes contained in the shapefile.  The missing value,
    resbdf, is assigned as the coordinates of a point separating
    different shapes.
    """
    result[:,:,:,:,:,:] = resbdf
    sf = shapefile.Reader(inputs[0])
    try:
        pt_index = 0
        for shp in sf.shapes():
            for pt in shp.points:
                result[pt_index,:2,0,0,0,0] = pt[:2]
                pt_index += 1
            # missing value coordinates (already assigned) separating shapes
            pt_index += 1
    except IndexError:
        # hit the maximum number of points
        pass


#
# The rest of this is for testing from the command line
#
if __name__ == "__main__":
    import time

    # make sure ferret_init does not cause problems
    info = ferret_init(0)

    resbdf = numpy.array([-9999.0], dtype=numpy.float64)
    inpbdfs = numpy.array([-8888.0, -7777.0], dtype=numpy.float64)
    maxpts = 3200 * 2400
    result = -6666.0 * numpy.ones((maxpts,2,1,1,1,1), dtype=numpy.float64, order='F')
    print("ferret_compute start: time = %s" % time.asctime())
    ferret_compute(0, result, resbdf, ("tl_2010_us_county10", maxpts, ), inpbdfs)
    print("ferret_compute done; time = %s" % time.asctime())
    good_x = numpy.logical_and((-180.0 <= result[:,0,0,0,0,0]), (result[:,0,0,0,0,0] <= -65.0))
    good_x = numpy.logical_or(good_x,
                 numpy.logical_and((172.0 <= result[:,0,0,0,0,0]), (result[:,0,0,0,0,0] <= 180.0)))
    good_y = numpy.logical_and((17.0 <= result[:,1,0,0,0,0]), (result[:,1,0,0,0,0] <= 72.0))
    if numpy.logical_xor(good_x, good_y).any():
        raise ValueError("good_x != good_y")
    missing_x = ( result[:,0,0,0,0,0] == resbdf )
    if numpy.logical_xor(good_x, numpy.logical_not(missing_x)).any():
        raise ValueError("good_x != not missing_x")
    missing_y = ( result[:,1,0,0,0,0] == resbdf )
    if numpy.logical_xor(good_y, numpy.logical_not(missing_y)).any():
        raise ValueError("good_y != not missing_y")
    count = 0
    at_end = False
    shape_num = 0
    total = 0
    for k in range(result.shape[0]):
        if missing_x[k]:
            if count == 0:
                at_end = True
            else:
                # print "Count[%d] = %d" % (shape_num, count)
                shape_num += 1
                total += count + 1
                count = 0
        elif at_end:
            raise ValueError("More than one missing value between shapes")
        else:
            count += 1
    total += count
    print("total (including missing-value separators) = %d" % total)
    print("out of a maximum of %d" %  result.shape[0])
    print("number of shapes = %d" % shape_num)
    print("shapefile_readxy: SUCCESS")




# To be run in python after importing and starting pyferret
# such as from running "pyferret -python"

from __future__ import print_function

import numpy

import sys ; sys.ps1 = '' ; sys.ps2 = ''
print()

print(">>> normax = pyferret.FerAxis()")
normax = pyferret.FerAxis()
print(">>> print repr(normax)")
print(repr(normax))
print(">>> dir(normax)")
dir(normax)

print(">>> coads = pyferret.FerDSet('coads_climatology')")
coads = pyferret.FerDSet('coads_climatology')
print(">>> coads.sst.load()")
coads.sst.load()
print(">>> sstaxes = coads.sst.grid.axes")
sstaxes = coads.sst.grid.axes
print(">>> print repr(sstaxes)")
print(repr(sstaxes))
print(">>> normax == sstaxes[0]")
normax == sstaxes[0]
print(">>> normax != sstaxes[1]")
normax != sstaxes[1]
print(">>> normax == sstaxes[2]")
normax == sstaxes[2]
print(">>> normax != sstaxes[3]")
normax != sstaxes[3]

print(">>> print repr(sstaxes[0].axtype)")
print(repr(sstaxes[0].axtype))
print(">>> print repr(sstaxes[0].coords)")
print(repr(sstaxes[0].coords))
print(">>> print repr(sstaxes[0].unit)")
print(repr(sstaxes[0].unit))
print(">>> print repr(sstaxes[0].name)")
print(repr(sstaxes[0].name))

print(">>> print repr(sstaxes[1].axtype)")
print(repr(sstaxes[1].axtype))
print(">>> print repr(sstaxes[1].coords)")
print(repr(sstaxes[1].coords))
print(">>> print repr(sstaxes[1].unit)")
print(repr(sstaxes[1].unit))
print(">>> print repr(sstaxes[1].name)")
print(repr(sstaxes[1].name))

print(">>> print repr(sstaxes[2].axtype)")
print(repr(sstaxes[2].axtype))
print(">>> print repr(sstaxes[2].coords)")
print(repr(sstaxes[2].coords))
print(">>> print repr(sstaxes[2].unit)")
print(repr(sstaxes[2].unit))
print(">>> print repr(sstaxes[2].name)")
print(repr(sstaxes[2].name))

print(">>> print repr(sstaxes[3].axtype)")
print(repr(sstaxes[3].axtype))
print(">>> print repr(sstaxes[3].coords)")
print(repr(sstaxes[3].coords))
print(">>> print repr(sstaxes[3].unit)")
print(repr(sstaxes[3].unit))
print(">>> print repr(sstaxes[3].name)")
print(repr(sstaxes[3].name))

print(">>> dupaxis = sstaxes[0].copy()")
dupaxis = sstaxes[0].copy()
print(">>> dupaxis is sstaxes[0]")
dupaxis is sstaxes[0]
print(">>> dupaxis == sstaxes[0]")
dupaxis == sstaxes[0]
print(">>> dupaxis.coords is sstaxes[0].coords")
dupaxis.coords is sstaxes[0].coords
print(">>> numpy.allclose(dupaxis.coords, sstaxes[0].coords)")
numpy.allclose(dupaxis.coords, sstaxes[0].coords)

print(">>> dupaxis = sstaxes[3].copy()")
dupaxis = sstaxes[3].copy()
print(">>> dupaxis is sstaxes[3]")
dupaxis is sstaxes[3]
print(">>> dupaxis == sstaxes[3]")
dupaxis == sstaxes[3]
print(">>> dupaxis.coords is sstaxes[3].coords")
dupaxis.coords is sstaxes[3].coords
print(">>> numpy.allclose(dupaxis.coords, sstaxes[3].coords)")
numpy.allclose(dupaxis.coords, sstaxes[3].coords)

print(">>> print repr(pyferret.FerAxis._parsegeoval(None))")
print(repr(pyferret.FerAxis._parsegeoval(None)))
print(">>> print repr(pyferret.FerAxis._parsegeoval(0))")
print(repr(pyferret.FerAxis._parsegeoval(0)))
print(">>> print repr(pyferret.FerAxis._parsegeoval(0.0))")
print(repr(pyferret.FerAxis._parsegeoval(0.0)))
print(">>> print repr(pyferret.FerAxis._parsegeoval('0'))")
print(repr(pyferret.FerAxis._parsegeoval('0')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('5E'))")
print(repr(pyferret.FerAxis._parsegeoval('5E')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('6W'))")
print(repr(pyferret.FerAxis._parsegeoval('6W')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('7N'))")
print(repr(pyferret.FerAxis._parsegeoval('7N')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('8S'))")
print(repr(pyferret.FerAxis._parsegeoval('8S')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('9m'))")
print(repr(pyferret.FerAxis._parsegeoval('9m')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('03-APR-2005 06:07:08'))")
print(repr(pyferret.FerAxis._parsegeoval('03-APR-2005 06:07:08')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('03-APR-2005 06:07'))")
print(repr(pyferret.FerAxis._parsegeoval('03-APR-2005 06:07')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('03-APR-2005'))")
print(repr(pyferret.FerAxis._parsegeoval('03-APR-2005')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('2003-04-05T06:07:08'))")
print(repr(pyferret.FerAxis._parsegeoval('2003-04-05T06:07:08')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('2003-04-05T06:07'))")
print(repr(pyferret.FerAxis._parsegeoval('2003-04-05T06:07')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('2003-04-05 06:07:08'))")
print(repr(pyferret.FerAxis._parsegeoval('2003-04-05 06:07:08')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('2003-04-05 06:07'))")
print(repr(pyferret.FerAxis._parsegeoval('2003-04-05 06:07')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('2003-04-05'))")
print(repr(pyferret.FerAxis._parsegeoval('2003-04-05')))
print(">>> print repr(pyferret.FerAxis._parsegeoval('4y', istimestep=True))")
print(repr(pyferret.FerAxis._parsegeoval('4y', istimestep=True)))
print(">>> print repr(pyferret.FerAxis._parsegeoval('6d', istimestep=True))")
print(repr(pyferret.FerAxis._parsegeoval('6d', istimestep=True)))
print(">>> print repr(pyferret.FerAxis._parsegeoval('7h', istimestep=True))")
print(repr(pyferret.FerAxis._parsegeoval('7h', istimestep=True)))
print(">>> print repr(pyferret.FerAxis._parsegeoval('8m', istimestep=True))")
print(repr(pyferret.FerAxis._parsegeoval('8m', istimestep=True)))
print(">>> print repr(pyferret.FerAxis._parsegeoval('9s', istimestep=True))")
print(repr(pyferret.FerAxis._parsegeoval('9s', istimestep=True)))
print(">>> print repr(pyferret.FerAxis._parsegeoval('1', istimestep=True))")
print(repr(pyferret.FerAxis._parsegeoval('1', istimestep=True)))


print(">>> print repr(pyferret.FerAxis._parsegeoslice( slice(5,23,2) ))")
print(repr(pyferret.FerAxis._parsegeoslice( slice(5,23,2) )))
print(">>> print repr(pyferret.FerAxis._parsegeoslice( slice(-5.0,15.0,4.0) ))")
print(repr(pyferret.FerAxis._parsegeoslice( slice(-5.0,15.0,4.0) )))
print(">>> print repr(pyferret.FerAxis._parsegeoslice( slice('-6','11','5') ))")
print(repr(pyferret.FerAxis._parsegeoslice( slice('-6','11','5') )))
print(">>> print repr(pyferret.FerAxis._parsegeoslice( slice('25W','35E',5) ))")
print(repr(pyferret.FerAxis._parsegeoslice( slice('25W','35E',5) )))
print(">>> print repr(pyferret.FerAxis._parsegeoslice( slice('15S','30N',3) ))")
print(repr(pyferret.FerAxis._parsegeoslice( slice('15S','30N',3) )))
print(">>> print repr(pyferret.FerAxis._parsegeoslice( slice('-900m','-100m','50m') ))")
print(repr(pyferret.FerAxis._parsegeoslice( slice('-900m','-100m','50m') )))
print(">>> print repr(pyferret.FerAxis._parsegeoslice( slice('03-APR-2005 11:30','23-JUL-2006 23:30','12h') ))")
print(repr(pyferret.FerAxis._parsegeoslice( slice('03-APR-2005 11:30','23-JUL-2006 23:30','12h') )))




'''
Maintains a dictionary of classes which can be used to bind PyFerret
graphics calls to a particular graphics engine.

When a new Window is needed, the createWindow function in this module
is called.  This function creates an instance of the appropriate
bindings class, its createWindow method is called, and the bindings 
instance is returned.

Other methods of the bindings instance are called directly to perform
graphics operations on this Window.
'''

from __future__ import print_function

from pyferret.graphbind.abstractpyferretbindings import AbstractPyFerretBindings

__pyferret_bindings_classes = { }

def addPyFerretBindings(engine_name, bindings_class):
    '''
    Adds the given class to the dictionary of available bindings.

    Arguments:
        engine_name: a string identifying the bindings class
                and graphics engine
        bindings_class: a subclass of AbstractPyFerretBindings
                providing the bindings for a graphics engine

    Raises:
        ValueError if bindings for engine_name already exist
        TypeError if bindings_class is not a subclass of
                AbstractPyFerretBindings
    '''
    try:
        _ = __pyferret_bindings_classes[engine_name]
        raise ValueError("Bindings already exist for graphics engine '%s'" \
                         % engine_name)
    except KeyError:
        pass
    if not issubclass(bindings_class, AbstractPyFerretBindings):
        raise TypeError("bindings_class argument is not a " \
                        "subclass of AbstractPyFerretBindings")
    __pyferret_bindings_classes[engine_name] = bindings_class

def knownPyFerretEngines():
    '''
    Returns a tuple of all the known engine names.
    '''
    return tuple( __pyferret_bindings_classes.keys() )
    
def createWindow(engine_name, title, visible, noalpha, rasteronly):
    '''
    Creates an instance of the bindings class associated with
    engine_name and calls the createWindow method of that
    instance of the class.  The instance of the bindings class
    is then returned.

    Arguments:
        engine_name: string identifying the bindings class and
                     graphics engine to use for the Window
        title: display title for the Window
        visible: display Window on start-up?
        noalpha: do not use the alpha channel in colors?
        rasteronly: only raster images will be used ?

    Returns the instance of the binding class associated with
    the newly created Window if the createWindow method of the
    instance returns True.  Otherwise the bindings instance is
    deleted and None is returned.

    Raises a ValueError if engine_name does not have any bindings.
    '''
    try:
        bindclass = __pyferret_bindings_classes[engine_name]
    except KeyError:
        raise ValueError("Unknown graphics engine '%s'" % engine_name)
    bindinst = bindclass()
    if not bindinst.createWindow(title, visible, noalpha, rasteronly):
        del bindinst
        return None
    return bindinst


if __name__ == "__main__":

    class TestBindings(AbstractPyFerretBindings):
        engine_name = "TestEngine"

        def __init__(self):
            super(TestBindings, self).__init__()

        def createWindow(self, title, visible, noalpha):
            return True

    addPyFerretBindings(TestBindings.engine_name, TestBindings)
    bindinst = createWindow(TestBindings.engine_name, "test", False, False, False)
    if not bindinst:
        raise RuntimeError("Unsuccessful creation of a Window")
    try:
        addPyFerretBindings(TestBindings.engine_name, TestBindings)
        raise RuntimeError("Adding bindings with the same name succeeded")
    except ValueError:
        pass
    try:
        addPyFerretBindings("GenericObject", object)
        raise RuntimeError("Adding object for a bindings class succeeded")
    except TypeError:
        pass
    known_engines = knownPyFerretEngines()
    if (len(known_engines) != 1) or \
       (known_engines[0] != TestBindings.engine_name):
        raise RuntimeError("Unexpected tuple of known engines: %s" % \
                           str(known_engines))
    print("Success")




from numpy.distutils.core import setup, Extension
import distutils.sysconfig
import sys
import os
import os.path
import re

# Get BUILDTYPE for checking if this is intel-mac
buildtype = os.getenv("BUILDTYPE")
if buildtype:
    buildtype = buildtype.strip()
if not buildtype:
    raise ValueError("Environment variable BUILDTYPE is not defined")

# (Non-standard) Directories containing .h include files
incdir_list = [ "pyfermod",
                os.path.join("fer", "common"),
                os.path.join("fmt", "cmn"),
                os.path.join("fer", "ef_utility"),
                os.path.join("fer", "grdel"), ]

bind_and_hide_internal = os.getenv("BIND_AND_HIDE_INTERNAL")
if bind_and_hide_internal:
    bind_and_hide_internal = bind_and_hide_internal.strip()

# NETCDF_LIBDIR must be given, either for the static library or the shared-object library
netcdf_libdir = os.getenv("NETCDF_LIBDIR")
if netcdf_libdir:
    netcdf_libdir = netcdf_libdir.strip()
if not netcdf_libdir:
    raise ValueError("Environment variable NETCDF_LIBDIR is not defined")

# HDF5_LIBDIR is only given if the HDF5 and NetCDF libraries are to be statically linked
hdf5_libdir = os.getenv("HDF5_LIBDIR")
if hdf5_libdir:
    hdf5_libdir = hdf5_libdir.strip()

# SZ_LIBDIR is the location of the SZ library to be linked in
sz_libdir = os.getenv("SZ_LIBDIR")
if sz_libdir:
    sz_libdir = sz_libdir.strip()

# CAIRO_LIBDIR is only given if the cairo library is to be statically linked in
cairo_libdir = os.getenv("CAIRO_LIBDIR")
if cairo_libdir:
    cairo_libdir = cairo_libdir.strip()

# PIXMAN_LIBDIR is only given if the pixman-1 library is to be statically linked in
pixman_libdir = os.getenv("PIXMAN_LIBDIR")
if pixman_libdir:
    pixman_libdir = pixman_libdir.strip()

# PANGO_LIBDIR gives a non-standard location of the pango libraries
pango_libdir = os.getenv("PANGO_LIBDIR")
if pango_libdir:
    pango_libdir = pango_libdir.strip()

# GFORTRAN_LIB gives a non-standard full-path location of the gfortran library to be used
# in the linking step.  If not given or empty, the -lgfortran flag is used in the linking step.
gfortran_lib = os.getenv("GFORTRAN_LIB")
if gfortran_lib:
    gfortran_lib = gfortran_lib.strip()

# The location of libpythonx.x.so, in case it is not in a standard location
python_libdir = os.path.split( distutils.sysconfig.get_python_lib(standard_lib=True) )[0]

# The list of additional directories to examine for libraries
libdir_list = [ "lib", netcdf_libdir, ]
if hdf5_libdir:
    libdir_list.append(hdf5_libdir)
if sz_libdir:
    libdir_list.append(sz_libdir)
if cairo_libdir:
    libdir_list.append(cairo_libdir)
if pixman_libdir:
    libdir_list.append(pixman_libdir)
if pango_libdir:
    libdir_list.append(pango_libdir)
libdir_list.append(python_libdir)

# Get the list of ferret static libraries
# Stripping off the "lib" prefix and the ".a" suffix
fer_lib_list = [ ]
for libname in os.listdir("lib"):
    if (libname[:3] == "lib") and (libname[-2:] == ".a"):
        fer_lib_list.append(libname[3:-2])

# Create the list of libraries to link
lib_list = fer_lib_list[:]
if buildtype != "intel-mac":
    # fer_lib_list is included multiple times to resolve interdependencies
    lib_list.extend(fer_lib_list)
    lib_list.extend(fer_lib_list)
    lib_list.extend(fer_lib_list)
    lib_list.extend(fer_lib_list)

# Linking in the rest of the system libraries were moved to addn_link_flags
# in order to make sure the appropriate netcdff, netcdf, hdf5_hl, hdf5, and
# cairo libraries are used.

addn_link_args = [ ]
# Link to the appropriate netcdf libraries.
# The hdf5 libraries are only used to resolve netcdf library function
# calls when statically linking in the netcdf libraries.
if hdf5_libdir:
    netcdff_lib = os.path.join(netcdf_libdir, "libnetcdff.a")
    addn_link_args.append(netcdff_lib)
    netcdf_lib = os.path.join(netcdf_libdir, "libnetcdf.a")
    addn_link_args.append(netcdf_lib)
    hdf5_hl_lib = os.path.join(hdf5_libdir, "libhdf5_hl.a")
    addn_link_args.append(hdf5_hl_lib)
    hdf5_lib = os.path.join(hdf5_libdir, "libhdf5.a")
    addn_link_args.append(hdf5_lib)
else:
    addn_link_args.extend([ "-lnetcdff", "-lnetcdf" ])

# Link to the cairo library and the libraries it requires.
if cairo_libdir:
    cairo_lib = os.path.join(cairo_libdir, "libcairo.a")
    addn_link_args.append(cairo_lib);
    if pixman_libdir:
        pixman_lib = os.path.join(pixman_libdir, "libpixman-1.a")
    else:
        pixman_lib = "-lpixman-1"
    addn_link_args.extend([ pixman_lib, "-lfreetype", "-lfontconfig", "-lpng" ])
else:
    addn_link_args.append("-lcairo")

# The Pango-Cairo text-rendering libraries
addn_link_args.append("-lpangocairo-1.0")

# Link in the appropriate system libraries
if hdf5_libdir:
    addn_link_args.append("-lcurl")
if sz_libdir:
    addn_link_args.append("-lsz")
if gfortran_lib:
    addn_link_args.append(gfortran_lib)
else:
    addn_link_args.append("-lgfortran")
addn_link_args.extend([ "-lz", "-ldl", "-lm", "-fPIC", ])

if bind_and_hide_internal:
    # Bind symbols and function symbols to any internal definitions
    # and do not make any of the symbols or function symbols defined
    # in any libraries externally visible (mainly for cairo and pixman).
    # Those in the object files (including those from pyfermod and
    # fer/ef_utility) will still be visible.
    addn_link_args.extend([ "-Wl,-Bsymbolic", "-Wl,--exclude-libs,ALL"])

if os.uname()[0] == 'Darwin':
    # For Mac OSX, leave room for library path renames
    addn_link_args.append("-Wl,-headerpad_max_install_names")

# Get the list of C source files in pyfermod
src_list = [ ]
for srcname in os.listdir("pyfermod"):
    if srcname[-2:] == ".c":
        src_list.append(os.path.join("pyfermod", srcname))

# Get the list of additional objects to be linked in
addnobjs_list = [ ]
dirname = os.path.join("fer", "ef_utility")
for srcname in os.listdir(dirname):
    if srcname[-2:] == ".o":
        addnobjs_list.append(os.path.join(dirname, srcname))
dirname = os.path.join("fer", "special")
for srcname in ( "FerMem_routines.o", "fakes3.o", "ferret_dispatch.o", "gui_fakes.o", "linux_routines.o", ):
    addnobjs_list.append(os.path.join(dirname, srcname))
for srcname in os.listdir(dirname):
    if (srcname[0] == 'x') and (srcname[-7:] == "_data.o"):
        addnobjs_list.append(os.path.join(dirname, srcname))

if bind_and_hide_internal:
    # Duplicate objects in libraries to make them externally visible (e.g., for las
    # external functions) if the '--exclude-libs ALL' flag was passed to the linker.
    dirname = os.path.join("fmt", "src")
    addnobjs_list.append(os.path.join(dirname, "tm_lenstr.o"));
    addnobjs_list.append(os.path.join(dirname, "tm_fmt.o"));
    addnobjs_list.append(os.path.join(dirname, "tm_lefint.o"));

# Create the pyferret.libpyferret Extension
ext_mods = [ Extension("pyferret.libpyferret", include_dirs = incdir_list,
                                               sources = src_list,
                                               extra_objects = addnobjs_list,
                                               library_dirs = libdir_list,
                                               libraries = lib_list,
                                               extra_link_args = addn_link_args), ]

pyferret_version = None
xrev_name = os.path.join("fer", "dat", "xrevision_data.F")
xrev_file = open(xrev_name)
try:
    pat = re.compile('\\s+DATA\\s+revision_level\\s*/\\s*(\\S+)\\s*/\\s*', flags=re.IGNORECASE)
    for datlin in xrev_file:
        mat = re.match(pat, datlin)
        if mat:
            pyferret_version = mat.group(1)
            break
finally:
    xrev_file.close()
if not pyferret_version:
    raise ValueError('Unable to find the version number in ' + xrev_name)

# Configure the setup
setup(name = "pyferret",
      version = pyferret_version,
      description = "Python module providing Ferret functionality",
      long_description = "Python module providing Ferret functionality",
      author = "Karl M. Smith",
      author_email = "karl.smith@noaa.gov",
      url = "http://ferret.pmel.noaa.gov/Ferret/documentation/pyferret",
      license = "Public Domain",
      requires = [ "numpy", ],
      packages = [ "pyferret", "pyferret.eofanal", "pyferret.fershp",
                   "pyferret.graphbind", "pyferret.regrid", "pyferret.stats", ],
      package_dir = { "pyferret":"pyfermod", },
      ext_modules = ext_mods)

setup(name = "pipedviewer",
      version = pyferret_version,
      description = "Graphics viewer controlled by a command pipe",
      long_description = "A graphics viewer application that receives its " \
                         "drawing and other commands primarily from another " \
                         "application through a pipe.  A limited number of " \
                         "commands are provided by the viewer itself to allow " \
                         "saving and some manipulation of the displayed scene.  " \
                         "The controlling application, however, will be unaware " \
                         "of these modifications made to the scene.",
      author = "Karl M. Smith",
      author_email = "karl.smith@noaa.gov",
      url = "http://ferret.pmel.noaa.gov/Ferret/documentation/pyferret",
      license = "Public Domain",
      requires = [ "multiprocessing", ],
      packages = [ "pipedviewer", ],
      package_dir = { "pipedviewer":"pviewmod", })

setup(name = "gcircle",
      version = pyferret_version,
      description = "Module of functions involving great circles with " \
                    "points given in longitudes and latitudes (thus " \
                    "assuming a spheroid model of the earth).",
      long_description = "Module of functions involving great circles with " \
                         "points given in longitudes and latitudes (thus " \
                         "assuming a spheroid model of the earth).",
      author = "Karl M. Smith",
      author_email = "karl.smith@noaa.gov",
      url = "http://ferret.pmel.noaa.gov/Ferret/documentation/pyferret",
      license = "Public Domain",
      requires = [ "numpy", ],
      py_modules = [ "gcircle", ])




'''
Regridder for converting data between a curvilinear longitude,
latitude, depth (or elevation) grid and a rectilinear longitude,
latitude, depth (or elevation) grid.  Uses the ESMP interface
to ESMF to perform the regridding.

@author: Karl Smith
'''

from __future__ import print_function

import numpy
import ESMP


class CurvRect3DRegridder(object):
    '''
    Regridder for regridding data between a 3D curvilinear grid, where
    the longitude, latitude, and depth (or elevation) of each grid corner
    and/or center point is explicitly defined, and a 3D rectilinear grid,
    where the grid corners are all intersections of sets of strictly
    increasing longitudes, strictly increasing latitudes, and strictly
    increasing depths (or elevations).  The rectilinear grid centers are
    the intersections of averaged consecutive longitude pairs, averaged
    consecutive latitude pairs, and averages consecutive depth (or
    elevation) pairs.

    For these grids, the center point [i, j, k] is taken to be the center
    point of the quadrilaterally-faced hexahedron defined by the corner
    points: corner_pt[i, j, k], corner_pt[i+1, j, k], corner_pt[i+1, j+1, k],
    corner_pt([i, j+1, k], corner_pt[i, j, k+1], corner_pt[i+1, j, k+1],
    cornmer_pt[i+1, j+1, k+1], and corner_pt[i, j+1, k+1].

    Uses the ESMP interface to ESMF to perform the regridding.  Cartesian
    grids are used in ESMP to perform this regridding, assuming a spherical
    earth with radius 6371.007 kilometers.

    Prior to calling any instance methods in the CurvRect3DRegridder class,
    the ESMP module must be imported and ESMP.ESMP_Initialize() must have
    been called.  When a CurvRect3DRegridder instance is no longer needed,
    the finalize method of the instance should be called to free ESMP
    resources associated with the instance.  When ESMP is no longer
    required, the ESMP.ESMP_Finalize() method should be called to free
    all ESMP and ESMF resources.

    See the ESMPControl singleton class to simplify initializing and
    finalizing ESMP once, and only once, for a Python session.
    '''


    def __init__(self, earth_radius=6371.007, debug=False):
        '''
        Initializes to an empty regridder.  The ESMP module must be
        imported and ESMP.ESMP_Initialize() called (possibly through
        invoking ESMPControl().startCheckESMP()) prior to calling
        any methods in this instance.

        Arguments:
            earth_rad_km: earth radius, in kilometers, to use in computing
                          cartesian coordinates for the regridding process.
                          The precision of this value is not critical and
                          the authalic radius given as the default should
                          suffice for most purposes.
        '''
        # Call the no-argument __init__ of the superclass
        super(CurvRect3DRegridder, self).__init__()
        # Radius of the earth, in kilometers, to use in regridding
        self.__earth_rad = earth_radius
        self.__debug = debug
        # tuples giving the shape of the grid (defined by number of cells)
        self.__curv_shape = None
        self.__rect_shape = None
        # ESMP_Grid objects describing the grids
        self.__curv_grid = None
        self.__rect_grid = None
        # ESMP_Field objects providing the source or regridded data
        self.__curv_dest_field = None
        self.__curv_src_field = None
        self.__rect_dest_field = None
        self.__rect_src_field = None
        # handles to regridding operations between the fields/grids
        self.__curv_to_rect_handles = { }
        self.__rect_to_curv_handles = { }


    def createCurvGrid(self, center_lons, center_lats, center_levs,
                       center_ignore=None, levels_are_depths=True,
                       corner_lons=None, corner_lats=None,
                       corner_levs=None, corner_ignore=None):
        '''
        Create the curvilinear grid as an ESMP_Grid using the provided center
        longitudes, latitudes, and depths (or elevations) as the grid
        center points, and, if given, the grid corner longitudes, latitudes,
        and depths (or elevations) as the grid corner points.  Curvilinear
        data is assigned to the center points.  Grid point coordinate
        coord[i,j,k] is assigned from lon[i,j,k], lat[i,j,k], and levs[i,j,k].

        Center point [i,j,k] is taken to be the center point of the
        quadrilaterally-faced hexahedron defined by the corner points:
        corner_pt[i, j, k], corner_pt[i+1, j, k], corner_pt[i+1, j+1, k],
        corner_pt([i, j+1, k], corner_pt[i, j, k+1], corner_pt[i+1, j, k+1],
        corner_pt[i+1, j+1, k+1], and corner_pt[i, j+1, k+1].

        Any previous ESMP_Grid, ESMP_Field, or ESMP regridding procedures
        are destroyed.

        Arguments:
            center_lons:       3D array of longitudes, in degrees,
                               for each of the curvilinear center points
            center_lats:       3D array of latitudes, in degrees,
                               for each of the curvilinear center points
            center_levs:       3D array of levels, in meters from the earth
                               radius, for each of the curvilinear center
                               points
            center_ignore:     3D array of boolean-like values, indicating
                               if the corresponding grid center point should
                               be ignored in the regridding; if None, no
                               grid center points will be ignored
            levels_are_depths: True -  levels are depths and subtracted from
                                       the earth radius;
                               False - levels are elevations and added to
                                       the earth radius
            corner_lons:       3D array of longitudes, in degrees,
                               for each of the curvilinear corner points
            corner_lats:       3D array of latitudes, in degrees,
                               for each of the curvilinear corner points
            corner_levs:       3D array of levels, in meters from the earth radius,
                               for each of the curvilinear corner points
            corner_ignore:     3D array of boolean-like values, indicating
                               if the corresponding grid corner point should
                               be ignored in the regridding; if None, no
                               grid corner points will be ignored
        Returns:
            None
        Raises:
            ValueError: if the shape (dimensionality) of an argument is
                        invalid, or if a value in an argument is invalid
            TypeError:  if an argument is not array-like
        '''
        # Make sure center_lons is an appropriate array-like argument
        center_lons_array = numpy.array(center_lons, dtype=numpy.float64, copy=False)
        if len(center_lons_array.shape) != 3:
            raise ValueError("center_lons must be three-dimensional")

        # Make sure center_lats is an appropriate array-like argument
        center_lats_array = numpy.array(center_lats, dtype=numpy.float64, copy=False)
        if center_lats_array.shape != center_lons_array.shape:
            raise ValueError("center_lats and center_lons must have the same shape")

        # Make sure center_lats is an appropriate array-like argument
        center_levs_array = numpy.array(center_levs, dtype=numpy.float64, copy=False)
        if center_levs_array.shape != center_lons_array.shape:
            raise ValueError("center_levs and center_lons must have the same shape")

        if center_ignore == None:
            # Using all points, no mask created
            center_ignore_array = None
        else:
            # Make sure ignore_pts is an appropriate array-like argument
            center_ignore_array = numpy.array(center_ignore, dtype=numpy.bool, copy=False)
            if center_ignore_array.shape != center_lons_array.shape:
                raise ValueError("center_ignore and center_lons must have the same shape")
            # If not actually ignoring any points, do not create a mask
            if not center_ignore_array.any():
                center_ignore_array = None

        # Mask for X,Y,Z coordinates to compute
        if center_ignore_array != None:
            good_centers = numpy.logical_not(center_ignore_array).flatten('F')
        else:
            good_centers = numpy.array([True] * (center_lons_array.shape[0] * \
                                                 center_lons_array.shape[1] * \
                                                 center_lons_array.shape[2]) )

        if (corner_lons != None) and (corner_lats != None) and (corner_levs != None):
            # Corner points specified

            # Make sure corner_lons is an appropriate array-like argument
            corner_lons_array = numpy.array(corner_lons, dtype=numpy.float64, copy=False)
            if corner_lons_array.shape != (center_lons_array.shape[0] + 1,
                                           center_lons_array.shape[1] + 1,
                                           center_lons_array.shape[2] + 1):
                raise ValueError("corner_lons must have one more point along " \
                                 "each dimension when compared to center_lons")

            # Make sure corner_lats is an appropriate array-like argument
            corner_lats_array = numpy.array(corner_lats, dtype=numpy.float64, copy=False)
            if corner_lats_array.shape != corner_lons_array.shape:
                raise ValueError("corner_lats and corner_lons must have the same shape")

            # Make sure corner_lats is an appropriate array-like argument
            corner_levs_array = numpy.array(corner_levs, dtype=numpy.float64, copy=False)
            if corner_levs_array.shape != corner_lons_array.shape:
                raise ValueError("corner_levs and corner_lons must have the same shape")

            if corner_ignore == None:
                # Using all points, no mask created
                corner_ignore_array = None
            else:
                # Make sure ignore_pts is an appropriate array-like argument
                corner_ignore_array = numpy.array(corner_ignore, dtype=numpy.bool, copy=False)
                if corner_ignore_array.shape != corner_lons_array.shape:
                    raise ValueError("corner_ignore and corner_lons must have the same shape")
                # If not actually ignoring any points, do not create a mask
                if not corner_ignore_array.any():
                    corner_ignore_array = None

            # Mask for X,Y,Z coordinates to compute
            if corner_ignore_array != None:
                good_corners = numpy.logical_not(corner_ignore_array).flatten('F')
            else:
                good_corners = numpy.array([True] * (corner_lons_array.shape[0] * \
                                                     corner_lons_array.shape[1] * \
                                                     corner_lons_array.shape[2]) )

        elif (corner_lons != None) or (corner_lats != None) or (corner_levs != None):
            raise ValueError("one or two, but not all three, of corner_lons, " \
                             "corner_lats, and corner_levs are given")
        elif corner_ignore != None:
            raise ValueError("corner_ignore given without corner_lons, " \
                             "corner_lats, and corner_levs")
        else:
            # No corner points specified
            corner_lons_array = None
            corner_lats_array = None
            corner_levs_array = None
            corner_ignore_array = None

        # Release any regridding procedures and clear the dictionaries
        for handle in self.__rect_to_curv_handles.values():
            ESMP.ESMP_FieldRegridRelease(handle)
        self.__rect_to_curv_handles.clear()
        for handle in self.__curv_to_rect_handles.values():
            ESMP.ESMP_FieldRegridRelease(handle)
        self.__curv_to_rect_handles.clear()
        # Destroy any curvilinear ESMP_Fields
        if self.__curv_src_field != None:
            ESMP.ESMP_FieldDestroy(self.__curv_src_field)
            self.__curv_src_field = None
        if self.__curv_dest_field != None:
            ESMP.ESMP_FieldDestroy(self.__curv_dest_field)
            self.__curv_dest_field = None
        # Destroy any previous curvilinear ESMP_Grid
        if self.__curv_grid != None:
            ESMP.ESMP_GridDestroy(self.__curv_grid);
            self.__curv_grid = None

        # Create the curvilinear 3D cartesian coordinates ESMP_Grid
        # using ESMP_GridCreateNoPeriDim for the typical case (not
        # the whole world) in Ferret.
        self.__curv_shape = center_lons_array.shape
        grid_shape = numpy.array(self.__curv_shape, dtype=numpy.int32)
        self.__curv_grid = ESMP.ESMP_GridCreateNoPeriDim(grid_shape,
                                         ESMP.ESMP_COORDSYS_CART,
                                         ESMP.ESMP_TYPEKIND_R8)

        if corner_lons_array != None:
            # Allocate space for the grid corner coordinates
            ESMP.ESMP_GridAddCoord(self.__curv_grid, ESMP.ESMP_STAGGERLOC_CORNER_VFACE)

            # Retrieve the grid corner coordinate arrays in the ESMP_Grid
            grid_x_coords = ESMP.ESMP_GridGetCoordPtr(self.__curv_grid, 0,
                                                      ESMP.ESMP_STAGGERLOC_CORNER_VFACE)
            grid_y_coords = ESMP.ESMP_GridGetCoordPtr(self.__curv_grid, 1,
                                                      ESMP.ESMP_STAGGERLOC_CORNER_VFACE)
            grid_z_coords = ESMP.ESMP_GridGetCoordPtr(self.__curv_grid, 2,
                                                      ESMP.ESMP_STAGGERLOC_CORNER_VFACE)

            # Assign the cartesian coordinates of the grid corners in the ESMP_Grid
            lons = corner_lons_array.flatten('F')
            lons = numpy.deg2rad(lons[good_corners])
            lats = corner_lats_array.flatten('F')
            lats = numpy.deg2rad(lats[good_corners])
            levs = corner_levs_array.flatten('F')
            levs = levs[good_corners] / 1000.0
            if levels_are_depths:
                levs *= -1.0
            levs += self.__earth_rad
            levs /= self.__earth_rad
            # XY plane through the prime meridian, Z toward central atlantic
            # 0 lon, 0 lat = (0, 0, R); 90 lon, 0 lat = (R, 0, 0); * lon, 90 lat = (0, R, 0)
            cos_lats = numpy.cos(lats)
            grid_x_coords[:] = 0.0
            grid_x_coords[good_corners] = levs * numpy.sin(lons) * cos_lats
            grid_y_coords[:] = 0.0
            grid_y_coords[good_corners] = levs * numpy.sin(lats)
            grid_z_coords[:] = 0.0
            grid_z_coords[good_corners] = levs * numpy.cos(lons) * cos_lats

            if self.__debug:
                fout = open("curv_corner_xyz.txt", "w")
                try:
                    print("curv_corner_x = %s" % \
                        self.__myArrayStr(grid_x_coords, corner_lons_array.shape), file=fout)
                    print("curv_corner_y = %s" % \
                        self.__myArrayStr(grid_y_coords, corner_lons_array.shape), file=fout)
                    print("curv_corner_z = %s" % \
                        self.__myArrayStr(grid_z_coords, corner_lons_array.shape), file=fout)
                finally:
                    fout.close()
            
            # Add a mask if not considering all the corner points
            if corner_ignore_array != None:
                # Allocate space for the grid corners mask
                ESMP.ESMP_GridAddItem(self.__curv_grid, ESMP.ESMP_GRIDITEM_MASK,
                                                        ESMP.ESMP_STAGGERLOC_CORNER_VFACE)
                # Retrieve the grid corners mask array in the ESMP_Grid
                ignore_mask = ESMP.ESMP_GridGetItem(self.__curv_grid,
                                                    ESMP.ESMP_GRIDITEM_MASK,
                                                    ESMP.ESMP_STAGGERLOC_CORNER_VFACE)
                # Assign the mask in the ESMP_Grid; 
                # False (turns into zero) means use the point;
                # True (turns into one) means ignore the point
                ignore_mask[:] = corner_ignore_array.flatten('F')

        # Allocate space for the grid center coordinates
        ESMP.ESMP_GridAddCoord(self.__curv_grid, ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)

        # Retrieve the grid center coordinate arrays in the ESMP_Grid
        grid_x_coords = ESMP.ESMP_GridGetCoordPtr(self.__curv_grid, 0,
                                                  ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)
        grid_y_coords = ESMP.ESMP_GridGetCoordPtr(self.__curv_grid, 1,
                                                  ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)
        grid_z_coords = ESMP.ESMP_GridGetCoordPtr(self.__curv_grid, 2,
                                                  ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)

        # Assign the cartesian coordinates of the grid corners in the ESMP_Grid
        lons = center_lons_array.flatten('F')
        lons = numpy.deg2rad(lons[good_centers])
        lats = center_lats_array.flatten('F')
        lats = numpy.deg2rad(lats[good_centers])
        levs = center_levs_array.flatten('F')
        levs = levs[good_centers] / 1000.0
        if levels_are_depths:
            levs *= -1.0
        levs += self.__earth_rad
        levs /= self.__earth_rad
        # XY plane through the prime meridian, Z toward central atlantic
        # 0 lon, 0 lat = (0, 0, R); 90 lon, 0 lat = (R, 0, 0); * lon, 90 lat = (0, R, 0)
        cos_lats = numpy.cos(lats)
        grid_x_coords[:] = 0.0
        grid_x_coords[good_centers] = levs * numpy.sin(lons) * cos_lats
        grid_y_coords[:] = 0.0
        grid_y_coords[good_centers] = levs * numpy.sin(lats)
        grid_z_coords[:] = 0.0
        grid_z_coords[good_centers] = levs * numpy.cos(lons) * cos_lats

        if self.__debug:
            fout = open("curv_center_xyz.txt", "w")
            try:
                print("curv_center_x = %s" % \
                    self.__myArrayStr(grid_x_coords, self.__curv_shape), file=fout)
                print("curv_center_y = %s" % \
                    self.__myArrayStr(grid_y_coords, self.__curv_shape), file=fout)
                print("curv_center_z = %s" % \
                    self.__myArrayStr(grid_z_coords, self.__curv_shape), file=fout)
            finally:
                fout.close()

        # Add a mask if not considering all the center points
        if center_ignore_array != None:
            # Allocate space for the grid centers mask
            ESMP.ESMP_GridAddItem(self.__curv_grid, ESMP.ESMP_GRIDITEM_MASK,
                                                    ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)
            # Retrieve the grid centers mask array in the ESMP_Grid
            ignore_mask = ESMP.ESMP_GridGetItem(self.__curv_grid,
                                                ESMP.ESMP_GRIDITEM_MASK,
                                                ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)
            # Assign the mask in the ESMP_Grid; 
            # False (turns into zero) means use the point;
            # True (turns into one) means ignore the point
            ignore_mask[:] = center_ignore_array.flatten('F')


    def assignCurvField(self, data=None):
        '''
        Possibly creates, and possible assigns, an appropriate curvilinear
        ESMP_Field located at the center points of the grid.  An ESMP_Field
        is created if and only if it is not already present; thus allowing
        the "weights" computed from a previous regridding to be reused.

        If data is not None, assigns the data values to the curvilinear
        source ESMP_Field, creating the ESMP_field if it does not exist.

        If data is None, only creates the curvilinear destination ESMP_Field
        if it does not exist.

        Arguments:
            data: 3D array of data values to be assigned in an
                  ESMP_Field for the grid center points; if None,
                  no data in any ESMP_Field is modified.
        Returns:
            None
        Raises:
            ValueError: if data is not None, and if the shape
                        (dimensionality) of data is invalid or
                        if a value in data is not numeric
            TypeError:  if data, if not None, is not array-like
        '''
        if data == None:

            # Create the curvilinear destination ESMP_Field if it does not exist
            if self.__curv_dest_field == None:
                self.__curv_dest_field = ESMP.ESMP_FieldCreateGrid(self.__curv_grid,
                                                        "curv_dest_field",
                                                        ESMP.ESMP_TYPEKIND_R8,
                                                        ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)

        else:

            # Make sure data is an appropriate array-like argument
            data_array = numpy.array(data, dtype=numpy.float64, copy=True)
            if data_array.shape != self.__curv_shape:
                raise ValueError("data must have the same shape " \
                                 "as the center points arrays")

            # Create the curvilinear source ESMP_Field if it does not exist
            if self.__curv_src_field == None:
                self.__curv_src_field = ESMP.ESMP_FieldCreateGrid(self.__curv_grid,
                                                       "curv_src_field",
                                                       ESMP.ESMP_TYPEKIND_R8,
                                                       ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)

            # Retrieve the field data array in the ESMP_Field
            field_ptr = ESMP.ESMP_FieldGetPtr(self.__curv_src_field)

            # Assign the data to the field array in the ESMP_Field;
            # flatten in the same order as the grid coordinates
            field_ptr[:] = data_array.flatten('F')

            if self.__debug:
                fout = open("curv_data.txt", "w")
                try:
                    print("curv_data = %s" % \
                        self.__myArrayStr(field_ptr, self.__curv_shape), file=fout)
                finally:
                    fout.close()


    def createRectGrid(self, center_lons, center_lats, center_levs,
                       center_ignore=None, levels_are_depths=True,
                       corner_lons=None, corner_lats=None, corner_levs=None,
                       corner_ignore=None):
        '''
        Create the rectilinear grid as an ESMP_Grid using the provided cell
        center and edge longitudes, latitudes, and levels to define the grid
        center and corner points.  Rectilinear data is assigned to the center
        points.  Grid point [i,j,k] is assigned from lons[i], lats[j], and
        levs[k].

        Center point [i,j,k] is taken to be the center point of the
        quadrilaterally-faced hexahedron defined by the corner points:
        corner_pt[i, j, k], corner_pt[i+1, j, k], corner_pt[i+1, j+1, k],
        corner_pt([i, j+1, k], corner_pt[i, j, k+1], corner_pt[i+1, j, k+1],
        corner_pt[i+1, j+1, k+1], and corner_pt[i, j+1, k+1].

        Any previous ESMP_Grid, ESMP_Field, or ESMP regridding procedures
        are destroyed.

        Arguments:
            center_lons:       1D array of longitudes, in degrees,
                               for each of the rectilinear grid cells center
                               planes
            center_lats:       1D array of latitudes, in degrees,
                               for each of the rectilinear grid cells center
                               planes
            center_levs:       1D array of levels, in meters from the earth
                               radius, for each of the rectilinear grid cells
                               center planes
            center_ignore:     3D array of boolean-like values, indicating if
                               the corresponding grid center point should be
                               ignored in the regridding; if None, no grid
                               center points will be ignored
            levels_are_depths: True -  levels are depths and subtracted from
                                       the earth radius,
                               False - levels are elevations and added to
                                       the earth radius
            corner_lons:       1D array of longitudes, in degrees,
                               for each of the rectilinear grid cells corner
                               planes
            corner_lats:       1D array of latitudes, in degrees,
                               for each of the rectilinear grid cells corner
                               planes
            corner_levs:       1D array of levels, in meters from the earth
                               radius, for each of the rectilinear grid cells
                               corner planes
            corner_ignore:     3D array of boolean-like values, indicating if
                               the corresponding grid corner point should be
                               ignored in the regridding; if None, no grid
                               corner points will be ignored
        Returns:
            None
        Raises:
            ValueError: if the shape (dimensionality) of an argument is
                        invalid, or if a value in an argument is invalid
            TypeError:  if an argument is not array-like
        '''
        # Make sure center_lons is an appropriate array-like argument
        center_lons_array = numpy.array(center_lons, dtype=numpy.float64, copy=False)
        if len(center_lons_array.shape) != 1:
            raise ValueError("center_lons must be one-dimensional")

        # Make sure center_lats is an appropriate array-like argument
        center_lats_array = numpy.array(center_lats, dtype=numpy.float64, copy=True)
        if len(center_lats_array.shape) != 1:
            raise ValueError("center_lats must be one-dimensional")

        # Make sure center_levs is an appropriate array-like argument
        center_levs_array = numpy.array(center_levs, dtype=numpy.float64, copy=True)
        if len(center_levs_array.shape) != 1:
            raise ValueError("center_levs must be one-dimensional")

        if center_ignore == None:
            # Using all center points, no mask created
            center_ignore_array = None
        else:
            # Make sure center_ignore is an appropriate array-like argument
            center_ignore_array = numpy.array(center_ignore, dtype=numpy.bool, copy=False)
            if center_ignore_array.shape != (center_lons_array.shape[0],
                                             center_lats_array.shape[0],
                                             center_levs_array.shape[0]):
                raise ValueError("center_ignore must have the shape " \
                                 "( len(center_lons), len(center_lats), len(center_levs) )")
            # If not actually ignoring any center points, do not create a mask
            if not center_ignore_array.any():
                center_ignore_array = None

        if (corner_lons != None) and (corner_lats != None) and (corner_levs != None):
            # Corner points specified

            # Make sure corner_lons is an appropriate array-like argument
            corner_lons_array = numpy.array(corner_lons, dtype=numpy.float64, copy=False)
            if corner_lons_array.shape != (center_lons_array.shape[0] + 1, ):
                raise ValueError("corner_lons must have shape ( len(center_lons) + 1, )")

            # Make sure corner_lats is an appropriate array-like argument
            corner_lats_array = numpy.array(corner_lats, dtype=numpy.float64, copy=True)
            if corner_lats_array.shape != (center_lats_array.shape[0] + 1, ):
                raise ValueError("corner_lats must have shape ( len(center_lats) + 1, )")

            # Make sure corner_levs is an appropriate array-like argument
            corner_levs_array = numpy.array(corner_levs, dtype=numpy.float64, copy=True)
            if corner_levs_array.shape != (center_levs_array.shape[0] + 1, ):
                raise ValueError("corner_levs must have shape ( len(center_levs) + 1, )")

            if corner_ignore == None:
                # Using all corner points, no mask created
                corner_ignore_array = None
            else:
                # Make sure corner_ignore is an appropriate array-like argument
                corner_ignore_array = numpy.array(corner_ignore, dtype=numpy.bool, copy=False)
                if corner_ignore_array.shape != (corner_lons_array.shape[0],
                                                 corner_lats_array.shape[0],
                                                 corner_levs_array.shape[0]):
                    raise ValueError("corner_ignore must have the shape " \
                                     "( len(corner_lons), len(corner_lats), len(corner_levs) )")
                # If not actually ignoring any points, do not create a mask
                if not corner_ignore_array.any():
                    corner_ignore_array = None

        elif (corner_lons != None) or (corner_lats != None) or (corner_levs != None):
            raise ValueError("one or two, but not all three, of corner_lons, " \
                             "corner_lats, and corner_levs are given")
        elif corner_ignore != None:
            raise ValueError("corner_ignore given without corner_lons, " \
                             "corner_lats, and corner_levs")
        else:
            # No corner points specified
            corner_lons_array = None
            corner_lats_array = None
            corner_levs_array = None
            corner_ignore_array = None

        # Release any regridding procedures and clear the dictionaries
        for handle in self.__rect_to_curv_handles.values():
            ESMP.ESMP_FieldRegridRelease(handle)
        self.__rect_to_curv_handles.clear()
        for handle in self.__curv_to_rect_handles.values():
            ESMP.ESMP_FieldRegridRelease(handle)
        self.__curv_to_rect_handles.clear()
        # Destroy any rectilinear ESMP_Fields
        if self.__rect_src_field != None:
            ESMP.ESMP_FieldDestroy(self.__rect_src_field)
            self.__rect_src_field = None
        if self.__rect_dest_field != None:
            ESMP.ESMP_FieldDestroy(self.__rect_dest_field)
            self.__rect_dest_field = None
        # Destroy any previous rectilinear ESMP_Grid
        if self.__rect_grid != None:
            ESMP.ESMP_GridDestroy(self.__rect_grid);
            self.__rect_grid = None

        # Create the rectilinear 3D cartesian coordinates ESMP_Grid
        # using ESMP_GridCreateNoPeriDim for the typical case (not
        # the whole world) in Ferret.
        self.__rect_shape = (center_lons_array.shape[0],
                             center_lats_array.shape[0],
                             center_levs_array.shape[0])
        grid_shape = numpy.array(self.__rect_shape, dtype=numpy.int32)
        self.__rect_grid = ESMP.ESMP_GridCreateNoPeriDim(grid_shape,
                                         ESMP.ESMP_COORDSYS_CART,
                                         ESMP.ESMP_TYPEKIND_R8)

        if corner_lons_array != None:
            # Allocate space for the grid corner coordinates
            ESMP.ESMP_GridAddCoord(self.__rect_grid, ESMP.ESMP_STAGGERLOC_CORNER_VFACE)

            # Retrieve the grid corner coordinate arrays in the ESMP_Grid
            grid_x_coords = ESMP.ESMP_GridGetCoordPtr(self.__rect_grid, 0,
                                                      ESMP.ESMP_STAGGERLOC_CORNER_VFACE)
            grid_y_coords = ESMP.ESMP_GridGetCoordPtr(self.__rect_grid, 1,
                                                      ESMP.ESMP_STAGGERLOC_CORNER_VFACE)
            grid_z_coords = ESMP.ESMP_GridGetCoordPtr(self.__rect_grid, 2,
                                                      ESMP.ESMP_STAGGERLOC_CORNER_VFACE)

            # Assign the cartesian coordinates of the grid corners in the ESMP_Grid
            # numpy.tile([a,b,c], 2) = [a,b,c,a,b,c]
            # numpy.repeat([a,b,c], 2) = [a,a,b,b,c,c]
            corner_shape = (corner_lons_array.shape[0],
                            corner_lats_array.shape[0],
                            corner_levs_array.shape[0])
            lons = numpy.deg2rad(corner_lons_array)
            cos_lons = numpy.tile(numpy.cos(lons),
                                  corner_shape[1] * corner_shape[2])
            sin_lons = numpy.tile(numpy.sin(lons),
                                  corner_shape[1] * corner_shape[2])
            lats = numpy.deg2rad(corner_lats_array)
            cos_lats = numpy.tile(numpy.repeat(numpy.cos(lats), corner_shape[0]),
                                  corner_shape[2])
            sin_lats = numpy.tile(numpy.repeat(numpy.sin(lats), corner_shape[0]),
                                  corner_shape[2])
            levs = corner_levs_array / 1000.0
            if levels_are_depths:
                levs *= -1.0
            levs += self.__earth_rad
            levs /= self.__earth_rad
            levs = numpy.repeat(levs, corner_shape[0] * corner_shape[1])
            # XY plane through the prime meridian, Z toward central atlantic
            # 0 lon, 0 lat = (0, 0, R); 90 lon, 0 lat = (R, 0, 0); * lon, 90 lat = (0, R, 0)
            grid_x_coords[:] = levs * sin_lons * cos_lats
            grid_y_coords[:] = levs * sin_lats
            grid_z_coords[:] = levs * cos_lons * cos_lats

            if self.__debug:
                fout = open("rect_corner_xyz.txt", "w")
                try:
                    print("rect_corner_x = %s" % \
                        self.__myArrayStr(grid_x_coords, corner_shape), file=fout)
                    print("rect_corner_y = %s" % \
                        self.__myArrayStr(grid_y_coords, corner_shape), file=fout)
                    print("rect_corner_z = %s" % \
                        self.__myArrayStr(grid_z_coords, corner_shape), file=fout)
                finally:
                    fout.close()

            # Add a mask if not considering all the corner points
            if (corner_ignore_array != None):
                # Allocate space for the grid corners mask
                ESMP.ESMP_GridAddItem(self.__rect_grid, ESMP.ESMP_GRIDITEM_MASK,
                                                        ESMP.ESMP_STAGGERLOC_CORNER_VFACE)
                # Retrieve the grid corners mask array in the ESMP_Grid
                ignore_mask = ESMP.ESMP_GridGetItem(self.__rect_grid,
                                                    ESMP.ESMP_GRIDITEM_MASK,
                                                    ESMP.ESMP_STAGGERLOC_CORNER_VFACE)
                # Assign the mask in the ESMP_Grid; 
                # False (turns into zero) means use the point;
                # True (turns into one) means ignore the point;
                # flatten in column-major (F) order to match lon-lat-lev assignment order
                ignore_mask[:] = corner_ignore_array.flatten('F')

        # Allocate space for the grid center coordinates
        ESMP.ESMP_GridAddCoord(self.__rect_grid, ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)

        # Retrieve the grid center coordinate arrays in the ESMP_Grid
        grid_x_coords = ESMP.ESMP_GridGetCoordPtr(self.__rect_grid, 0,
                                                  ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)
        grid_y_coords = ESMP.ESMP_GridGetCoordPtr(self.__rect_grid, 1,
                                                  ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)
        grid_z_coords = ESMP.ESMP_GridGetCoordPtr(self.__rect_grid, 2,
                                                  ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)

        # Assign the cartesian coordinates of the grid centers in the ESMP_Grid
        # numpy.tile([a,b,c], 2) = [a,b,c,a,b,c]
        # numpy.repeat([a,b,c], 2) = [a,a,b,b,c,c]
        lons = numpy.deg2rad(center_lons_array)
        cos_lons = numpy.tile(numpy.cos(lons),
                              center_lats_array.shape[0] * center_levs_array.shape[0])
        sin_lons = numpy.tile(numpy.sin(lons),
                              center_lats_array.shape[0] * center_levs_array.shape[0])
        lats = numpy.deg2rad(center_lats_array)
        cos_lats = numpy.tile(numpy.repeat(numpy.cos(lats),
                                           center_lons_array.shape[0]),
                              center_levs_array.shape[0])
        sin_lats = numpy.tile(numpy.repeat(numpy.sin(lats),
                                           center_lons_array.shape[0]),
                              center_levs_array.shape[0])
        levs = center_levs_array / 1000.0
        if levels_are_depths:
            levs *= -1.0
        levs += self.__earth_rad
        levs /= self.__earth_rad
        levs = numpy.repeat(levs, center_lons_array.shape[0] * center_lats_array.shape[0])
        # XY plane through the prime meridian, Z toward central atlantic
        # 0 lon, 0 lat = (0, 0, R); 90 lon, 0 lat = (R, 0, 0); * lon, 90 lat = (0, R, 0)
        grid_x_coords[:] = levs * sin_lons * cos_lats
        grid_y_coords[:] = levs * sin_lats
        grid_z_coords[:] = levs * cos_lons * cos_lats

        if self.__debug:
            try:
                fout = open("rect_center_xyz.txt", "w")
                print("rect_center_x = %s" % \
                    self.__myArrayStr(grid_x_coords, self.__rect_shape), file=fout)
                print("rect_center_y = %s" % \
                    self.__myArrayStr(grid_y_coords, self.__rect_shape), file=fout)
                print("rect_center_z = %s" % \
                    self.__myArrayStr(grid_z_coords, self.__rect_shape), file=fout)
            finally:
                fout.close()

        # Add a mask if not considering all the center points
        if (center_ignore_array != None):
            # Allocate space for the grid centers mask
            ESMP.ESMP_GridAddItem(self.__rect_grid, ESMP.ESMP_GRIDITEM_MASK,
                                                    ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)
            # Retrieve the grid centers mask array in the ESMP_Grid
            ignore_mask = ESMP.ESMP_GridGetItem(self.__rect_grid,
                                                ESMP.ESMP_GRIDITEM_MASK,
                                                ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)
            # Assign the mask in the ESMP_Grid; 
            # False (turns into zero) means use the point;
            # True (turns into one) means ignore the point;
            # flatten in column-major (F) order to match lon-lat-lev assignment order
            ignore_mask[:] = center_ignore_array.flatten('F')


    def assignRectField(self, data=None):
        '''
        Possibly creates, and possible assigns, an appropriate rectilinear
        ESMP_Field located at the center points of the grid.  An ESMP_Field
        is created if and only if it is not already present; thus allowing
        the "weights" computed from a previous regridding to be reused.

        If data is not None, assigns the data values to the rectilinear
        source ESMP_Field, creating the ESMP_field if it does not exist.

        If data is None, only creates the rectilinear destination ESMP_Field
        if it does not exist.

        Arguments:
            data: 3D array of data values to be assigned in an
                  ESMP_Field for the grid center points; if None,
                  no data in any ESMP_Field is modified.
        Returns:
            None
        Raises:
            ValueError: if data is not None, and if the shape
                        (dimensionality) of data is invalid or
                        if a value in data is not numeric
            TypeError:  if data, if not None, is not array-like
        '''
        if data == None:

            # Create the rectilinear destination ESMP_Field if it does not exist
            if self.__rect_dest_field == None:
                self.__rect_dest_field = ESMP.ESMP_FieldCreateGrid(self.__rect_grid,
                                                        "rect_dest_field",
                                                        ESMP.ESMP_TYPEKIND_R8,
                                                        ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)

        else:

            # Make sure data is an appropriate array-like argument
            data_array = numpy.array(data, dtype=numpy.float64, copy=True)
            if data_array.shape != self.__rect_shape:
                raise ValueError("data must have the shape " \
                                 "( len(center_lats), len(center_lats), len(center_levs) ) ")

            # Create the rectilinear source ESMP_Field if it does not exist
            if self.__rect_src_field == None:
                self.__rect_src_field = ESMP.ESMP_FieldCreateGrid(self.__rect_grid,
                                                       "rect_src_field",
                                                       ESMP.ESMP_TYPEKIND_R8,
                                                       ESMP.ESMP_STAGGERLOC_CENTER_VCENTER)

            # Retrieve the field data array in the ESMP_Field
            field_ptr = ESMP.ESMP_FieldGetPtr(self.__rect_src_field)

            # Assign the data to the field array in the ESMP_Field;
            # flatten in column-major (F) order to match lon-lat-lev assignment order
            field_ptr[:] = data_array.flatten('F')
            if self.__debug:
                fout = open("rect_data.txt", "w")
                try:
                    print("rect_data = %s" % \
                        self.__myArrayStr(field_ptr, self.__rect_shape), file=fout)
                finally:
                    fout.close()


    def regridCurvToRect(self, undef_val,
                         method=ESMP.ESMP_REGRIDMETHOD_BILINEAR):
        '''
        Regrids from the curvilinear source ESMP_Field to the rectilinear
        destination ESMP_Field using the given regridding method.  Reuses
        the appropriate regridding procedure if one already exists;
        otherwise a new regridding procedure is created and stored.

        Prior to calling this method, the curvilinear source ESMP_Field
        must be created by calling createCurvGrid, then assignCurvField
        with valid data.  The rectilinear destination ESMP_Field must
        also have been created by calling createRectGrid, and then
        assignRectField with no data argument (or None for data).

        Arguments:
            undef_val: numpy array containing one numeric value to
                       be used as the undefined data value in the
                       returned array
            method:    one of the ESMP regridding method identifiers,
                       such as:
                           ESMP.ESMP_REGRIDMETHOD_BILINEAR
                           ESMP.ESMP_REGRIDMETHOD_CONSERVE
                           ESMP.ESMP_REGRIDMETHOD_PATCH
                       Conservative regridding requires that both
                       corner and center point coordinates are
                       defined in the grids.
        Returns:
            data: a 3D numpy array of data values located at the
                  rectilinear grid centers representing the regridded
                  curvilinear ESMP_Field data.  The undefined data
                  value will be assigned to unassigned data points.
        Raises:
            ValueError: if either the curvilinear source ESMP_Field
                        or the rectilinear destination ESMP_Field
                        does not exist.
        '''
        # Check that the source and destination fields exist
        if self.__curv_src_field == None:
            raise ValueError("Curvilinear source ESMP_Field does not exist")
        if self.__rect_dest_field == None:
            raise ValueError("Rectilinear destination ESMP_Field does not exist")

        # Check if a regrid procedure handle already exists for this method
        handle = self.__curv_to_rect_handles.get(method, None)
        # If no handle found, create one
        if handle == None:
            # Assign the value in the masks marking points to be ignored
            ignore_mask_value = numpy.array([1], dtype=numpy.int32)
            # Generate the procedure handle
            handle = ESMP.ESMP_FieldRegridStore(self.__curv_src_field,
                                                self.__rect_dest_field,
                                                ignore_mask_value, ignore_mask_value,
                                                method, ESMP.ESMP_UNMAPPEDACTION_IGNORE)
            # Save the handle for this method for future regrids
            self.__curv_to_rect_handles[method] = handle

        # Initialize the destination field values with the undefined data value
        field_ptr = ESMP.ESMP_FieldGetPtr(self.__rect_dest_field)
        field_ptr[:] = undef_val

        # Perform the regridding, zeroing out only the 
        # destination fields values that will be assigned
        ESMP.ESMP_FieldRegrid(self.__curv_src_field, self.__rect_dest_field,
                              handle, ESMP.ESMP_REGION_SELECT)

        # Make a copy of the destination field values to return, reshaped to 3D
        result = numpy.array(field_ptr, dtype=numpy.float64, copy=True)
        if self.__debug:
            fout = open("regr_rect_data.txt", "w")
            try:
                print("regr_rect_data = %s" % \
                    self.__myArrayStr(result, self.__rect_shape), file=fout)
            finally:
                fout.close()

        result = result.reshape(self.__rect_shape, order='F')

        return result


    def regridRectToCurv(self, undef_val,
                         method=ESMP.ESMP_REGRIDMETHOD_BILINEAR):
        '''
        Regrids from the rectilinear source ESMP_Field to the curvilinear
        destination ESMP_Field using the given regridding method.  Reuses
        the appropriate regridding procedure if one already exists;
        otherwise a new regridding procedure is created and stored.

        Prior to calling this method, the rectilinear source ESMP_Field
        must be created by calling createRectGrid, then assignRectField
        with valid data.  The curvilinear destination ESMP_Field must
        also have been created by calling createCurvGrid, and then
        assignCurvField with no data argument (or None for data).

        Arguments:
            undef_val: numpy array containing one numeric value to
                       be used as the undefined data value in the
                       returned array
            method:    one of the ESMP regridding method identifiers,
                       such as:
                           ESMP.ESMP_REGRIDMETHOD_BILINEAR
                           ESMP.ESMP_REGRIDMETHOD_CONSERVE
                           ESMP.ESMP_REGRIDMETHOD_PATCH
                       Conservative regridding requires that both
                       corner and center point coordinates are
                       defined in the grids.
        Returns:
            data: a 3D array of data values located at the curvilinear
                  grid centers representing the regridded rectilinear
                  ESMP_Field data.  The undefined data value will be
                  assigned to unassigned data points.
        Raises:
            ValueError: if either the rectilinear source ESMP_Field
                        or the curvilinear destination ESMP_Field
                        does not exist.
        '''
        # Check that the source and destination fields exist
        if self.__rect_src_field == None:
            raise ValueError("Rectilinear source ESMP_Field does not exist")
        if self.__curv_dest_field == None:
            raise ValueError("Curvilinear destination ESMP_Field does not exist")

        # Check if a regrid procedure handle already exists for this method
        handle = self.__rect_to_curv_handles.get(method, None)
        # If no handle found, create one
        if handle == None:
            # Assign the value in the masks marking points to be ignored
            ignore_mask_value = numpy.array([1], dtype=numpy.int32)
            # Generate the procedure handle
            handle = ESMP.ESMP_FieldRegridStore(self.__rect_src_field,
                                                self.__curv_dest_field,
                                                ignore_mask_value, ignore_mask_value,
                                                method, ESMP.ESMP_UNMAPPEDACTION_IGNORE)
            # Save the handle for this method for future regrids
            self.__rect_to_curv_handles[method] = handle

        # Initialize the destination field values with the undefined data value
        field_ptr = ESMP.ESMP_FieldGetPtr(self.__curv_dest_field)
        field_ptr[:] = undef_val

        # Perform the regridding, zeroing out only the 
        # destination fields values that will be assigned
        ESMP.ESMP_FieldRegrid(self.__rect_src_field, self.__curv_dest_field,
                              handle, ESMP.ESMP_REGION_SELECT)

        # Make a copy of the destination field values to return, reshaped to 3D
        result = numpy.array(field_ptr, dtype=numpy.float64, copy=True)
        if self.__debug:
            fout = open("regr_curv_data.txt", "w")
            try:
                print("regr_curv_data = %s" % \
                    self.__myArrayStr(result, self.__curv_shape), file=fout)
            finally:
                fout.close()
        result = result.reshape(self.__curv_shape, order='F')

        return result


    def finalize(self):
        '''
        Destroys any ESMP_Grid, ESMP_Field, and ESMP regridding
        procedures present in this instance.  If ESMP is no longer
        needed, ESMP.ESMP_Finalize() should be called to free any
        ESMP and ESMF resources.

        Arguments:
            None
        Returns:
            None
        '''
        # Release any regridding procedures and clear the dictionaries
        for handle in self.__rect_to_curv_handles.values():
            ESMP.ESMP_FieldRegridRelease(handle)
        self.__rect_to_curv_handles.clear()
        for handle in self.__curv_to_rect_handles.values():
            ESMP.ESMP_FieldRegridRelease(handle)
        self.__curv_to_rect_handles.clear()
        # Destroy any ESMP_Fields
        if self.__rect_src_field != None:
            ESMP.ESMP_FieldDestroy(self.__rect_src_field)
            self.__rect_src_field = None
        if self.__rect_dest_field != None:
            ESMP.ESMP_FieldDestroy(self.__rect_dest_field)
            self.__rect_dest_field = None
        if self.__curv_src_field != None:
            ESMP.ESMP_FieldDestroy(self.__curv_src_field)
            self.__curv_src_field = None
        if self.__curv_dest_field != None:
            ESMP.ESMP_FieldDestroy(self.__curv_dest_field)
            self.__curv_dest_field = None
        # Destroy any ESMP_Grids
        if self.__rect_grid != None:
            ESMP.ESMP_GridDestroy(self.__rect_grid);
            self.__rect_grid = None
        if self.__curv_grid != None:
            ESMP.ESMP_GridDestroy(self.__curv_grid);
            self.__curv_grid = None
        # clear the shape attributes just to be complete
        self.__rect_shape = None
        self.__curv_shape = None


    def __myArrayStr(self, arr, shape):
        '''
        Private utility method for debug printing of data arrays

        Arguments:
            arr -   numpy array (3D array flattened in Fortran order)
                    to be printed
            shape - tuple giving the 3D shape for the arr
        Returns:
            string representation of arr as
                "numpy.array([...]).reshape((...), order='F')"
            with newlines providing a block structure of the data.
            Data values are printed in format %#10.6f
        '''
        mystr = 'numpy.array([\n'
        k = 0
        for val in arr:
            mystr += '%#10.6f,' % val
            k += 1
            if k % shape[0] == 0:
                mystr += '\n'
            if k % (shape[0] * shape[1]) == 0:
                mystr += '\n'
        if mystr[-2:] == '\n\n':
            mystr = mystr[:-1] + ']'
        elif mystr[-1] == '\n':
            mystr += ']'
        else:
            mystr += '\n]'
        mystr += ").reshape(%s, order='F')" % str(shape)
        return mystr

